"""
Tests the financial data structures
"""

import unittest
import os
import numpy as np
import pandas as pd

from famlafl.data_structures import run_data_structures as ds


class TestDataStructures(unittest.TestCase):
    """
    Test the various financial data structures:
    1. Run Dollar bars
    2. Run Volume bars
    3. Run Tick bars
    """

    def setUp(self):
        """
        Set the file path for the tick data csv
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/imbalance_sample_data_small.csv'

    def test_ema_run_dollar_bars(self):
        """
        Tests the EMA run dollar bars implementation.
        """
        exp_num_ticks_init = 1000
        num_prev_bars = 3

        db1, thresh_1 = ds.get_ema_dollar_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                   expected_imbalance_window=10000,
                                                   num_prev_bars=num_prev_bars, batch_size=2e7, verbose=False,
                                                   analyse_thresholds=True)
        db2, thresh_2 = ds.get_ema_dollar_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                   expected_imbalance_window=10000,
                                                   num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                                   analyse_thresholds=True)
        db3, _ = ds.get_ema_dollar_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                            expected_imbalance_window=10000,
                                            num_prev_bars=num_prev_bars, batch_size=10, verbose=False)
        ds.get_ema_dollar_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                   expected_imbalance_window=10000,
                                   num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                   to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (2, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        self.assertTrue(np.all(thresh_1.cum_theta_buy == thresh_2.cum_theta_buy))
        self.assertTrue(np.all(thresh_1.cum_theta_sell == thresh_2.cum_theta_sell))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1303.00)
        self.assertEqual(db1.loc[0, 'close'], 1305.75)

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[1, 'open'], 1305.75)
        self.assertEqual(db1.loc[1, 'high'], 1308.75)
        self.assertEqual(db1.loc[1, 'low'], 1305.25)
        self.assertEqual(db1.loc[1, 'close'], 1307.25)

        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())
        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_ema_run_volume_bars(self):
        """
        Tests the EMA run volume bars implementation.
        """
        exp_num_ticks_init = 1000
        num_prev_bars = 3

        db1, thresh_1 = ds.get_ema_volume_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                   expected_imbalance_window=10000,
                                                   num_prev_bars=num_prev_bars, batch_size=2e7, verbose=False,
                                                   analyse_thresholds=True)
        db2, thresh_2 = ds.get_ema_volume_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                   expected_imbalance_window=10000,
                                                   num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                                   analyse_thresholds=True)
        db3, _ = ds.get_ema_volume_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                            expected_imbalance_window=10000,
                                            num_prev_bars=num_prev_bars, batch_size=10, verbose=False)
        ds.get_ema_volume_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                   expected_imbalance_window=10000,
                                   num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                   to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (2, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        self.assertTrue(np.all(thresh_1.cum_theta_buy == thresh_2.cum_theta_buy))
        self.assertTrue(np.all(thresh_1.cum_theta_sell == thresh_2.cum_theta_sell))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1303.00)
        self.assertEqual(db1.loc[0, 'close'], 1305.75)

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[1, 'open'], 1305.75)
        self.assertEqual(db1.loc[1, 'high'], 1308.75)
        self.assertEqual(db1.loc[1, 'low'], 1305.25)
        self.assertEqual(db1.loc[1, 'close'], 1307.25)

        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())
        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_ema_run_tick_bars(self):
        """
        Tests the EMA run tick bars implementation.
        """
        exp_num_ticks_init = 1000
        num_prev_bars = 3

        db1, thresh_1 = ds.get_ema_tick_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                 expected_imbalance_window=10000,
                                                 num_prev_bars=num_prev_bars, batch_size=2e7, verbose=False,
                                                 analyse_thresholds=True)
        db2, thresh_2 = ds.get_ema_tick_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                 expected_imbalance_window=10000,
                                                 num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                                 analyse_thresholds=True)
        db3, _ = ds.get_ema_tick_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                          expected_imbalance_window=10000,
                                          num_prev_bars=num_prev_bars, batch_size=10, verbose=False)
        ds.get_ema_tick_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                 expected_imbalance_window=10000,
                                 num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                 to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (2, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        self.assertTrue(np.all(thresh_1.cum_theta_buy == thresh_2.cum_theta_buy))
        self.assertTrue(np.all(thresh_1.cum_theta_sell == thresh_2.cum_theta_sell))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1303.00)
        self.assertEqual(db1.loc[0, 'close'], 1305.75)

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[1, 'open'], 1305.75)
        self.assertEqual(db1.loc[1, 'high'], 1308.75)
        self.assertEqual(db1.loc[1, 'low'], 1305.25)
        self.assertEqual(db1.loc[1, 'close'], 1307.25)

        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())
        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_ema_run_dollar_bars_with_constraints(self):
        """
        Test the EMA Dollar Run bars with expected number of ticks max and min constraints
        """
        exp_num_ticks_init = 1000
        num_prev_bars = 3
        exp_num_ticks_constraints = [100, 1000]

        db1, _ = ds.get_ema_dollar_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                            expected_imbalance_window=10000,
                                            exp_num_ticks_constraints=exp_num_ticks_constraints,
                                            num_prev_bars=num_prev_bars, batch_size=2e7, verbose=False)
        db2, _ = ds.get_ema_dollar_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                            expected_imbalance_window=10000,
                                            exp_num_ticks_constraints=exp_num_ticks_constraints,
                                            num_prev_bars=num_prev_bars, batch_size=50, verbose=False)
        db3, _ = ds.get_ema_dollar_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                            expected_imbalance_window=10000,
                                            exp_num_ticks_constraints=exp_num_ticks_constraints,
                                            num_prev_bars=num_prev_bars, batch_size=10, verbose=False)
        ds.get_ema_dollar_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                   expected_imbalance_window=10000,
                                   exp_num_ticks_constraints=exp_num_ticks_constraints,
                                   num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                   to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (5, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1303.0)
        self.assertEqual(db1.loc[0, 'close'], 1305.75)

        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())

        # Assert OHLC is correct (some index)
        self.assertEqual(db1.loc[4, 'open'], 1306)
        self.assertEqual(db1.loc[4, 'high'], 1306.75)
        self.assertEqual(db1.loc[4, 'low'], 1303.5)
        self.assertEqual(db1.loc[4, 'close'], 1303.5)

        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_const_run_dollar_bars(self):
        """
        Tests the Const run dollar bars implementation.
        """
        exp_num_ticks_init = 1000
        num_prev_bars = 3

        db1, thresh_1 = ds.get_const_dollar_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                     expected_imbalance_window=10000,
                                                     num_prev_bars=num_prev_bars, batch_size=2e7, verbose=False,
                                                     analyse_thresholds=True)
        db2, thresh_2 = ds.get_const_dollar_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                     expected_imbalance_window=10000,
                                                     num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                                     analyse_thresholds=True)
        db3, _ = ds.get_const_dollar_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                              expected_imbalance_window=10000,
                                              num_prev_bars=num_prev_bars, batch_size=10, verbose=False)
        ds.get_const_dollar_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                     expected_imbalance_window=10000,
                                     num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                     to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (5, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        self.assertTrue(np.all(thresh_1.cum_theta_buy == thresh_2.cum_theta_buy))
        self.assertTrue(np.all(thresh_1.cum_theta_sell == thresh_2.cum_theta_sell))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1303.00)
        self.assertEqual(db1.loc[0, 'close'], 1305.75)

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[4, 'open'], 1306)
        self.assertEqual(db1.loc[4, 'high'], 1306.75)
        self.assertEqual(db1.loc[4, 'low'], 1303.5)
        self.assertEqual(db1.loc[4, 'close'], 1303.5)

        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())
        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_const_run_volume_bars(self):
        """
        Tests the Const run volume bars implementation.
        """
        exp_num_ticks_init = 1000
        num_prev_bars = 3

        db1, thresh_1 = ds.get_const_volume_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                     expected_imbalance_window=10000,
                                                     num_prev_bars=num_prev_bars, batch_size=2e7, verbose=False,
                                                     analyse_thresholds=True)
        db2, thresh_2 = ds.get_const_volume_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                     expected_imbalance_window=10000,
                                                     num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                                     analyse_thresholds=True)
        db3, _ = ds.get_const_volume_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                              expected_imbalance_window=10000,
                                              num_prev_bars=num_prev_bars, batch_size=10, verbose=False)
        ds.get_const_volume_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                     expected_imbalance_window=10000,
                                     num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                     to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (5, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        self.assertTrue(np.all(thresh_1.cum_theta_buy == thresh_2.cum_theta_buy))
        self.assertTrue(np.all(thresh_1.cum_theta_sell == thresh_2.cum_theta_sell))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1303.00)
        self.assertEqual(db1.loc[0, 'close'], 1305.75)

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[2, 'open'], 1306.0)
        self.assertEqual(db1.loc[2, 'high'], 1307.75)
        self.assertEqual(db1.loc[2, 'low'], 1305.75)
        self.assertEqual(db1.loc[2, 'close'], 1307.75)

        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())
        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_const_run_tick_bars(self):
        """
        Tests the Const run dollar bars implementation.
        """
        exp_num_ticks_init = 1000
        num_prev_bars = 3

        db1, thresh_1 = ds.get_const_tick_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                   expected_imbalance_window=10000,
                                                   num_prev_bars=num_prev_bars, batch_size=2e7, verbose=False,
                                                   analyse_thresholds=True)
        db2, thresh_2 = ds.get_const_tick_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                   expected_imbalance_window=10000,
                                                   num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                                   analyse_thresholds=True)
        db3, _ = ds.get_const_tick_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                            expected_imbalance_window=10000,
                                            num_prev_bars=num_prev_bars, batch_size=10, verbose=False)
        ds.get_const_tick_run_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                   expected_imbalance_window=10000,
                                   num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                   to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (5, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        self.assertTrue(np.all(thresh_1.cum_theta_buy == thresh_2.cum_theta_buy))
        self.assertTrue(np.all(thresh_1.cum_theta_sell == thresh_2.cum_theta_sell))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1303.00)
        self.assertEqual(db1.loc[0, 'close'], 1305.75)

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[2, 'open'], 1306.0)
        self.assertEqual(db1.loc[2, 'high'], 1307.5)
        self.assertEqual(db1.loc[2, 'low'], 1305.75)
        self.assertEqual(db1.loc[2, 'close'], 1307.5)

        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())
        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_csv_format(self):
        """
        Asserts that the csv data being passed is of the correct format.
        """
        wrong_date = ['2019-41-30', 200.00, np.int64(5)]
        wrong_price = ['2019-01-30', 'asd', np.int64(5)]
        wrong_volume = ['2019-01-30', 200.00, '1.5']
        too_many_cols = ['2019-01-30', 200.00,
                         np.int64(5), 'Limit order', 'B23']

        # pylint: disable=protected-access
        self.assertRaises(ValueError, ds.BaseRunBars._assert_csv,
                          pd.DataFrame(wrong_date).T)
        # pylint: disable=protected-access
        self.assertRaises(AssertionError,
                          ds.BaseRunBars._assert_csv,
                          pd.DataFrame(too_many_cols).T)
        # pylint: disable=protected-access
        self.assertRaises(AssertionError,
                          ds.BaseRunBars._assert_csv,
                          pd.DataFrame(wrong_price).T)
        # pylint: disable=protected-access
        self.assertRaises(AssertionError,
                          ds.BaseRunBars._assert_csv,
                          pd.DataFrame(wrong_volume).T)



"""
Tests ewma function from fast_ewma module
"""

import unittest
import os
import numpy as np
import pandas as pd

from famlafl.util.fast_ewma import ewma


class TestDataStructures(unittest.TestCase):
    """
    Test the various financial data structures:
    1. Imbalance Dollar bars
    2. Imbalance Volume bars
    3. Imbalance Tick bars
    """

    def setUp(self):
        """
        Set the file path for the tick data csv
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/tick_data.csv'

    def test_ewma(self):
        """
        Tests the imbalance dollar bars implementation.
        """
        test_sample = pd.read_csv(self.path)
        price_arr = np.array(test_sample.Price.values, dtype=float)
        ewma_res = ewma(price_arr, window=20)

        # Assert output array length equals input array length
        self.assertTrue(ewma_res.shape == price_arr.shape)
        # Assert the first value of ewma equals to input array value
        self.assertTrue(ewma_res[0] == price_arr[0])
        # Assert next value check with tolerance of 1e-5
        self.assertTrue(abs(ewma_res[1] - 1100.00) < 1e-5)



"""
Implementation of Trend-Scanning labels described in `Advances in Financial Machine Learning: Lecture 3/10
<https://papers.ssrn.com/sol3/papers.cfm?abstract_id=2708678>`_
"""

import pandas as pd
import numpy as np

from famlafl.structural_breaks.sadf import get_betas


def trend_scanning_labels(price_series: pd.Series, t_events: list = None, look_forward_window: int = 20,
                          min_sample_length: int = 5, step: int = 1) -> pd.DataFrame:
    """
    `Trend scanning <https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3257419>`_ is both a classification and
    regression labeling technique.

    That can be used in the following ways:

    1. Classification: By taking the sign of t-value for a given observation we can set {-1, 1} labels to define the
       trends as either downward or upward.
    2. Classification: By adding a minimum t-value threshold you can generate {-1, 0, 1} labels for downward, no-trend,
       upward.
    3. The t-values can be used as sample weights in classification problems.
    4. Regression: The t-values can be used in a regression setting to determine the magnitude of the trend.

    The output of this algorithm is a DataFrame with t1 (time stamp for the farthest observation), t-value, returns for
    the trend, and bin.

    :param price_series: (pd.Series) Close prices used to label the data set
    :param t_events: (list) Filtered events, array of pd.Timestamps
    :param look_forward_window: (int) Maximum look forward window used to get the trend value
    :param min_sample_length: (int) Minimum sample length used to fit regression
    :param step: (int) Optimal t-value index is searched every 'step' indices
    :return: (pd.DataFrame) Consists of t1, t-value, ret, bin (label information). t1 - label endtime, tvalue,
        ret - price change %, bin - label value based on price change sign
    """
    # pylint: disable=invalid-name

    if t_events is None:
        t_events = price_series.index

    t1_array = []  # Array of label end times
    t_values_array = []  # Array of trend t-values

    for index in t_events:
        subset = price_series.loc[index:].iloc[:look_forward_window]  # Take t:t+L window
        if subset.shape[0] >= look_forward_window:
            # Loop over possible look-ahead windows to get the one which yields maximum t values for b_1 regression coef
            max_abs_t_value = -np.inf  # Maximum abs t-value of b_1 coefficient among l values
            max_t_value_index = None  # Index with maximum t-value
            max_t_value = None  # Maximum t-value signed

            # Get optimal label end time value based on regression t-statistics
            for forward_window in np.arange(min_sample_length, subset.shape[0], step):
                y_subset = subset.iloc[:forward_window].values.reshape(-1, 1)  # y{t}:y_{t+l}

                # Array of [1, 0], [1, 1], [1, 2], ... [1, l] # b_0, b_1 coefficients
                X_subset = np.ones((y_subset.shape[0], 2))
                X_subset[:, 1] = np.arange(y_subset.shape[0])

                # Get regression coefficients estimates
                b_mean_, b_std_ = get_betas(X_subset, y_subset)
                # Check if l gives the maximum t-value among all values {0...L}
                t_beta_1 = (b_mean_[1] / np.sqrt(b_std_[1, 1]))[0]
                if abs(t_beta_1) > max_abs_t_value:
                    max_abs_t_value = abs(t_beta_1)
                    max_t_value = t_beta_1
                    max_t_value_index = forward_window

            # Store label information (t1, return)
            label_endtime_index = subset.index[max_t_value_index - 1]
            t1_array.append(label_endtime_index)
            t_values_array.append(max_t_value)

        else:
            t1_array.append(None)
            t_values_array.append(None)

    labels = pd.DataFrame({'t1': t1_array, 't_value': t_values_array}, index=t_events)
    labels.loc[:, 'ret'] = price_series.reindex(labels.t1).values / price_series.reindex(labels.index).values - 1
    labels['bin'] = labels.t_value.apply(np.sign)

    return labels



"""
Various volatility estimators
"""
import pandas as pd
import numpy as np


# pylint: disable=redefined-builtin

def get_daily_vol(close, lookback=100):
    """
    Advances in Financial Machine Learning, Snippet 3.1, page 44.

    Daily Volatility Estimates

    Computes the daily volatility at intraday estimation points.

    In practice we want to set profit taking and stop-loss limits that are a function of the risks involved
    in a bet. Otherwise, sometimes we will be aiming too high (tao ≫ sigma_t_i,0), and sometimes too low
    (tao ≪ sigma_t_i,0 ), considering the prevailing volatility. Snippet 3.1 computes the daily volatility
    at intraday estimation points, applying a span of lookback days to an exponentially weighted moving
    standard deviation.

    See the pandas documentation for details on the pandas.Series.ewm function.
    Note: This function is used to compute dynamic thresholds for profit taking and stop loss limits.

    :param close: (pd.Series) Closing prices
    :param lookback: (int) Lookback period to compute volatility
    :return: (pd.Series) Daily volatility value
    """
    # daily vol re-indexed to close
    df0 = close.index.searchsorted(close.index - pd.Timedelta(days=1))
    df0 = df0[df0 > 0]
    df0 = (pd.Series(close.index[df0 - 1], index=close.index[close.shape[0] - df0.shape[0]:]))

    df0 = close.loc[df0.index] / close.loc[df0.array].array - 1  # daily returns
    df0 = df0.ewm(span=lookback).std()
    return df0


def get_parksinson_vol(high: pd.Series, low: pd.Series, window: int = 20) -> pd.Series:
    """
    Parkinson volatility estimator

    :param high: (pd.Series): High prices
    :param low: (pd.Series): Low prices
    :param window: (int): Window used for estimation
    :return: (pd.Series): Parkinson volatility
    """
    ret = np.log(high / low)  # High/Low return
    estimator = 1 / (4 * np.log(2)) * (ret ** 2)
    return np.sqrt(estimator.rolling(window=window).mean())


def get_garman_class_vol(open: pd.Series, high: pd.Series, low: pd.Series, close: pd.Series,
                         window: int = 20) -> pd.Series:
    """
    Garman-Class volatility estimator

    :param open: (pd.Series): Open prices
    :param high: (pd.Series): High prices
    :param low: (pd.Series): Low prices
    :param close: (pd.Series): Close prices
    :param window: (int): Window used for estimation
    :return: (pd.Series): Garman-Class volatility
    """
    ret = np.log(high / low)  # High/Low return
    close_open_ret = np.log(close / open)  # Close/Open return
    estimator = 0.5 * ret ** 2 - (2 * np.log(2) - 1) * close_open_ret ** 2
    return np.sqrt(estimator.rolling(window=window).mean())


def get_yang_zhang_vol(open: pd.Series, high: pd.Series, low: pd.Series, close: pd.Series,
                       window: int = 20) -> pd.Series:
    """

    Yang-Zhang volatility estimator

    :param open: (pd.Series): Open prices
    :param high: (pd.Series): High prices
    :param low: (pd.Series): Low prices
    :param close: (pd.Series): Close prices
    :param window: (int): Window used for estimation
    :return: (pd.Series): Yang-Zhang volatility
    """
    k = 0.34 / (1.34 + (window + 1) / (window - 1))

    open_prev_close_ret = np.log(open / close.shift(1))
    close_prev_open_ret = np.log(close / open.shift(1))

    high_close_ret = np.log(high / close)
    high_open_ret = np.log(high / open)
    low_close_ret = np.log(low / close)
    low_open_ret = np.log(low / open)

    sigma_open_sq = 1 / (window - 1) * (open_prev_close_ret ** 2).rolling(window=window).sum()
    sigma_close_sq = 1 / (window - 1) * (close_prev_open_ret ** 2).rolling(window=window).sum()
    sigma_rs_sq = 1 / (window - 1) * (high_close_ret * high_open_ret + low_close_ret * low_open_ret).rolling(
        window=window).sum()

    return np.sqrt(sigma_open_sq + k * sigma_close_sq + (1 - k) * sigma_rs_sq)



# pylint: disable=missing-module-docstring

import unittest
import os
import numpy as np
import pandas as pd
from famlafl.labeling.fixed_time_horizon import fixed_time_horizon


class TestLabelingFixedTime(unittest.TestCase):
    """
    Tests regarding fixed time horizon labeling method.
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/stock_prices.csv'
        self.data = pd.read_csv(self.path, index_col='Date')
        self.data.index = pd.to_datetime(self.data.index)
        self.idx10 = self.data[:10].index

    def test_basic(self):
        """
        Tests for basic case, constant threshold and no standardization, lag.
        """
        close = self.data[['SPY', 'EPP', 'FXI']][:10]
        test1 = fixed_time_horizon(close['SPY'], lag=False)
        test2 = fixed_time_horizon(close, lag=False)
        test3 = fixed_time_horizon(close, lag=True)
        test4 = fixed_time_horizon(close, threshold=0.01, lag=True)
        test5 = fixed_time_horizon(close['SPY'], threshold=0.99, lag=False)
        test1_actual = pd.Series([np.nan, -1, -1, -1, -1, 1, 1, -1, 1, -1], index=self.idx10)
        test2_actual = pd.DataFrame({'SPY': [np.nan, -1, -1, -1, -1, 1, 1, -1, 1, -1],
                                     'EPP': [np.nan, 1, -1, 1, -1, 1, 1, -1, 1, -1],
                                     'FXI': [np.nan, -1, -1, 1, -1, 1, 1, -1, 1, -1]}, index=self.idx10)
        test3_actual = pd.DataFrame({'SPY': [-1, -1, -1, -1, 1, 1, -1, 1, -1, np.nan],
                                     'EPP': [1, -1, 1, -1, 1, 1, -1, 1, -1, np.nan],
                                     'FXI': [-1, -1, 1, -1, 1, 1, -1, 1, -1, np.nan]}, index=self.idx10)
        test4_actual = pd.DataFrame({'SPY': [0, -1, 0, -1, 1, 0, 0, 0, -1, np.nan],
                                     'EPP': [0, -1, 1, -1, 1, 1, -1, 1, -1, np.nan],
                                     'FXI': [0, -1, 1, -1, 1, 0, -1, 0, -1, np.nan]}, index=self.idx10)
        test5_actual = pd.Series([np.nan, 0, 0, 0, 0, 0, 0, 0, 0, 0], index=self.idx10)

        pd.testing.assert_series_equal(test1_actual, test1, check_names=False)
        pd.testing.assert_frame_equal(test2_actual, test2)
        pd.testing.assert_frame_equal(test3_actual, test3)
        pd.testing.assert_frame_equal(test4_actual, test4)
        pd.testing.assert_series_equal(test5_actual, test5, check_names=False)

    def test_dynamic_threshold(self):
        """
        Tests for when threshold is a pd.Series rather than a constant.
        """
        close = self.data[['SPY', 'EPP', 'FXI']][:10]
        threshold1 = pd.Series([0.01, 0.005, 0, 0.01, 0.02, 0.03, 0.1, -1, 0.99, 0], index=self.idx10)
        test6 = fixed_time_horizon(close, threshold=threshold1, lag=True)
        test7 = fixed_time_horizon(close['SPY'], threshold=threshold1, lag=False)
        test6_actual = pd.DataFrame({'SPY': [0, -1, -1, -1, 0, 0, 0, 1, 0, np.nan],
                                     'EPP': [0, -1, 1, -1, 0, 0, 0, 1, 0, np.nan],
                                     'FXI': [0, -1, 1, -1, 1, 0, 0, 1, 0, np.nan]}, index=self.idx10)
        test7_actual = pd.Series([np.nan, 0, -1, 0, 0, 0, 0, 1, 0, -1], index=self.idx10)

        pd.testing.assert_frame_equal(test6_actual, test6)
        pd.testing.assert_series_equal(test7_actual, test7, check_names=False)

    def test_with_standardization(self):
        """
        Test cases with standardization, with constant and dynamic threshold.
        """
        close = self.data[['SPY', 'EPP', 'FXI']][:10]
        threshold2 = pd.Series([1, 2, 0, 0.2, 0.02, 1.5, 10, -1, 500, 1], index=self.idx10)

        test8 = fixed_time_horizon(close, threshold=1, lag=False, standardized=True, window=4)
        test9 = fixed_time_horizon(close, threshold=0.1, lag=True, standardized=True, window=5)
        test10 = fixed_time_horizon(close, threshold=threshold2, lag=True, standardized=True, window=3)
        test8_actual = pd.DataFrame({'SPY': [np.nan, np.nan, np.nan, np.nan, 0, 1, 0, 0, 0, -1],
                                     'EPP': [np.nan, np.nan, np.nan, np.nan, 0, 1, 0, -1, 0, -1],
                                     'FXI': [np.nan, np.nan, np.nan, np.nan, 0, 1, 0, -1, 0, -1]},
                                    index=self.idx10)
        test9_actual = pd.DataFrame({'SPY': [np.nan, np.nan, np.nan, np.nan, 1, 1, -1, 1, -1, np.nan],
                                     'EPP': [np.nan, np.nan, np.nan, np.nan, 1, 1, -1, 1, -1, np.nan],
                                     'FXI': [np.nan, np.nan, np.nan, np.nan, 1, -1, -1, 0, -1, np.nan]},
                                    index=self.idx10)
        test10_actual = pd.DataFrame({'SPY': [np.nan, np.nan, 1, 0, 1, 0, 0, 1, 0, np.nan],
                                      'EPP': [np.nan, np.nan, 1, -1, 1, 0, 0, 1, 0, np.nan],
                                      'FXI': [np.nan, np.nan, 1, -1, 1, 0, 0, 1, 0, np.nan]},
                                     index=self.idx10)

        pd.testing.assert_frame_equal(test8_actual, test8)
        pd.testing.assert_frame_equal(test9_actual, test9)
        pd.testing.assert_frame_equal(test10_actual, test10)

    def test_resample(self):
        """
        Tests for when a resample period is used.
        """
        cols = ['SPY', 'EPP', 'FXI']
        close1 = self.data[cols].iloc[0:30]
        close2 = self.data[cols].iloc[0:150]
        week_index = close1.resample('W').last().index
        month_index = close2.resample('M').last().index
        threshold3 = pd.Series([0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1, 1.1], index=month_index)

        test11 = fixed_time_horizon(close1, threshold=0.02, resample_by='W', lag=True, standardized=False)
        test12 = fixed_time_horizon(close2, threshold=threshold3, resample_by='M', lag=True, standardized=True, window=3)
        test11_actual = pd.DataFrame({'SPY': [0, -1, 0, 1, -1, 1, np.nan],
                                      'EPP': [0, -1, 0, 1, -1, 0, np.nan],
                                      'FXI': [1, -1, -1, 0, -1, 1, np.nan]},
                                     index=week_index)
        test12_actual = pd.DataFrame({'SPY': [np.nan, np.nan, 1, 0, -1, 0, 0, np.nan],
                                      'EPP': [np.nan, np.nan, 1, 0, -1, 0, 0, np.nan],
                                      'FXI': [np.nan, np.nan, 1, 0, -1, 1, 0, np.nan]},
                                     index=month_index)

        pd.testing.assert_frame_equal(test11_actual, test11)
        pd.testing.assert_frame_equal(test12_actual, test12)

    def test_exceptions_warnings(self):
        """
        Tests the exceptions and warning that can be raised.
        """
        close = self.data[['SPY', 'EWG']][:10]
        threshold = pd.Series([0.01]*10)
        with self.assertRaises(Exception):  # Threshold index doesn't match.
            fixed_time_horizon(close, threshold)
        with self.assertRaises(Exception):  # Standardized but no window.
            fixed_time_horizon(close, 0.01, standardized=True)
        with self.assertWarns(UserWarning):  # Window too long.
            fixed_time_horizon(close, 0.01, standardized=True, window=50)



"""
First generation features (Roll Measure/Impact, Corwin-Schultz spread estimator)
"""

import numpy as np
import pandas as pd


def get_roll_measure(close_prices: pd.Series, window: int = 20) -> pd.Series:
    """
    Advances in Financial Machine Learning, page 282.

    Get Roll Measure

    Roll Measure gives the estimate of effective bid-ask spread
    without using quote-data.

    :param close_prices: (pd.Series) Close prices
    :param window: (int) Estimation window
    :return: (pd.Series) Roll measure
    """
    price_diff = close_prices.diff()
    price_diff_lag = price_diff.shift(1)
    return 2 * np.sqrt(abs(price_diff.rolling(window=window).cov(price_diff_lag)))


def get_roll_impact(close_prices: pd.Series, dollar_volume: pd.Series, window: int = 20) -> pd.Series:
    """
    Get Roll Impact.

    Derivate from Roll Measure which takes into account dollar volume traded.

    :param close_prices: (pd.Series) Close prices
    :param dollar_volume: (pd.Series) Dollar volume series
    :param window: (int) Estimation window
    :return: (pd.Series) Roll impact
    """
    roll_measure = get_roll_measure(close_prices, window)
    return roll_measure / dollar_volume


# Corwin-Schultz algorithm
def _get_beta(high: pd.Series, low: pd.Series, window: int) -> pd.Series:
    """
    Advances in Financial Machine Learning, Snippet 19.1, page 285.

    Get beta estimate from Corwin-Schultz algorithm

    :param high: (pd.Series) High prices
    :param low: (pd.Series) Low prices
    :param window: (int) Estimation window
    :return: (pd.Series) Beta estimates
    """
    ret = np.log(high / low)
    high_low_ret = ret ** 2
    beta = high_low_ret.rolling(window=2).sum()
    beta = beta.rolling(window=window).mean()
    return beta


def _get_gamma(high: pd.Series, low: pd.Series) -> pd.Series:
    """
    Advances in Financial Machine Learning, Snippet 19.1, page 285.

    Get gamma estimate from Corwin-Schultz algorithm.

    :param high: (pd.Series) High prices
    :param low: (pd.Series) Low prices
    :return: (pd.Series) Gamma estimates
    """
    high_max = high.rolling(window=2).max()
    low_min = low.rolling(window=2).min()
    gamma = np.log(high_max / low_min) ** 2
    return gamma


def _get_alpha(beta: pd.Series, gamma: pd.Series) -> pd.Series:
    """
    Advances in Financial Machine Learning, Snippet 19.1, page 285.

    Get alpha from Corwin-Schultz algorithm.

    :param beta: (pd.Series) Beta estimates
    :param gamma: (pd.Series) Gamma estimates
    :return: (pd.Series) Alphas
    """
    den = 3 - 2 * 2 ** .5
    alpha = (2 ** .5 - 1) * (beta ** .5) / den
    alpha -= (gamma / den) ** .5
    alpha[alpha < 0] = 0  # Set negative alphas to 0 (see p.727 of paper)
    return alpha


def get_corwin_schultz_estimator(high: pd.Series, low: pd.Series, window: int = 20) -> pd.Series:
    """
    Advances in Financial Machine Learning, Snippet 19.1, page 285.

    Get Corwin-Schultz spread estimator using high-low prices

    :param high: (pd.Series) High prices
    :param low: (pd.Series) Low prices
    :param window: (int) Estimation window
    :return: (pd.Series) Corwin-Schultz spread estimators
    """
    # Note: S<0 iif alpha<0
    beta = _get_beta(high, low, window)
    gamma = _get_gamma(high, low)
    alpha = _get_alpha(beta, gamma)
    spread = 2 * (np.exp(alpha) - 1) / (1 + np.exp(alpha))
    start_time = pd.Series(high.index[0:spread.shape[0]], index=spread.index)
    spread = pd.concat([spread, start_time], axis=1)
    spread.columns = ['Spread', 'Start_Time']  # 1st loc used to compute beta
    return spread.Spread


def get_bekker_parkinson_vol(high: pd.Series, low: pd.Series, window: int = 20) -> pd.Series:
    """
    Advances in Financial Machine Learning, Snippet 19.2, page 286.

    Get Bekker-Parkinson volatility from gamma and beta in Corwin-Schultz algorithm.

    :param high: (pd.Series) High prices
    :param low: (pd.Series) Low prices
    :param window: (int) Estimation window
    :return: (pd.Series) Bekker-Parkinson volatility estimates
    """
    # pylint: disable=invalid-name
    beta = _get_beta(high, low, window)
    gamma = _get_gamma(high, low)

    k2 = (8 / np.pi) ** 0.5
    den = 3 - 2 * 2 ** .5
    sigma = (2 ** -0.5 - 1) * beta ** 0.5 / (k2 * den)
    sigma += (gamma / (k2 ** 2 * den)) ** 0.5
    sigma[sigma < 0] = 0
    return sigma



"""
Logic regarding labeling from chapter 3. In particular the Triple Barrier Method and Meta-Labeling.
"""

import numpy as np
import pandas as pd

# Snippet 3.1, page 44, Daily Volatility Estimates
from famlafl.util.multiprocess import mp_pandas_obj


# Snippet 3.2, page 45, Triple Barrier Labeling Method
def apply_pt_sl_on_t1(close, events, pt_sl, molecule, **kwargs):  # pragma: no cover
    """
    Advances in Financial Machine Learning, Snippet 3.2, page 45.

    Triple Barrier Labeling Method

    This function applies the triple-barrier labeling method. It works on a set of
    datetime index values (molecule). This allows the program to parallelize the processing.

    Mainly it returns a DataFrame of timestamps regarding the time when the first barriers were reached.

    :param close: (pd.Series) Close prices
    :param events: (pd.Series) Indices that signify "events" (see cusum_filter function
    for more details)
    :param pt_sl: (np.array) Element 0, indicates the profit taking level; Element 1 is stop loss level
    :param molecule: (an array) A set of datetime index values for processing
    :param kwargs: Additional keyword arguments (ignored)
    :return: (pd.DataFrame) Timestamps of when first barrier was touched
    """
    # Apply stop loss/profit taking, if it takes place before t1 (end of event)
    events_ = events.loc[molecule]
    out = events_[['t1']].copy(deep=True)

    profit_taking_multiple = pt_sl[0]
    stop_loss_multiple = pt_sl[1]

    # Profit taking active
    if profit_taking_multiple > 0:
        profit_taking = profit_taking_multiple * events_['trgt']
    else:
        profit_taking = pd.Series(index=events.index)  # NaNs

    # Stop loss active
    if stop_loss_multiple > 0:
        stop_loss = -stop_loss_multiple * events_['trgt']
    else:
        stop_loss = pd.Series(index=events.index)  # NaNs

    out['pt'] = pd.Series(dtype=events.index.dtype)
    out['sl'] = pd.Series(dtype=events.index.dtype)

    # Get events
    for loc, vertical_barrier in events_['t1'].fillna(close.index[-1]).items():
        closing_prices = close[loc: vertical_barrier]  # Path prices for a given trade
        cum_returns = (closing_prices / close[loc] - 1) * events_.at[loc, 'side']  # Path returns
        out.at[loc, 'sl'] = cum_returns[cum_returns < stop_loss[loc]].index.min()  # Earliest stop loss date
        out.at[loc, 'pt'] = cum_returns[cum_returns > profit_taking[loc]].index.min()  # Earliest profit taking date

    return out


# Snippet 3.4 page 49, Adding a Vertical Barrier
def add_vertical_barrier(t_events, close, num_days=0, num_hours=0, num_minutes=0, num_seconds=0):
    """
    Advances in Financial Machine Learning, Snippet 3.4 page 49.

    Adding a Vertical Barrier

    For each index in t_events, it finds the timestamp of the next price bar at or immediately after
    a number of days num_days. This vertical barrier can be passed as an optional argument t1 in get_events.

    This function creates a series that has all the timestamps of when the vertical barrier would be reached.

    :param t_events: (pd.Series) Series of events (symmetric CUSUM filter)
    :param close: (pd.Series) Close prices
    :param num_days: (int) Number of days to add for vertical barrier
    :param num_hours: (int) Number of hours to add for vertical barrier
    :param num_minutes: (int) Number of minutes to add for vertical barrier
    :param num_seconds: (int) Number of seconds to add for vertical barrier
    :return: (pd.Series) Timestamps of vertical barriers
    """
    timedelta = pd.Timedelta(
        '{} days, {} hours, {} minutes, {} seconds'.format(num_days, num_hours, num_minutes, num_seconds))
    # Find index to closest to vertical barrier
    nearest_index = close.index.searchsorted(t_events + timedelta)

    # Exclude indexes which are outside the range of close price index
    nearest_index = nearest_index[nearest_index < close.shape[0]]

    # Find price index closest to vertical barrier time stamp
    nearest_timestamp = close.index[nearest_index]
    filtered_events = t_events[:nearest_index.shape[0]]

    vertical_barriers = pd.Series(data=nearest_timestamp, index=filtered_events)
    return vertical_barriers


# Snippet 3.3 -> 3.6 page 50, Getting the Time of the First Touch, with Meta Labels
def get_events(close, t_events, pt_sl, target, min_ret, num_threads, vertical_barrier_times=False,
               side_prediction=None, verbose=True):
    """
    Advances in Financial Machine Learning, Snippet 3.6 page 50.

    Getting the Time of the First Touch, with Meta Labels

    This function is orchestrator to meta-label the data, in conjunction with the Triple Barrier Method.

    :param close: (pd.Series) Close prices
    :param t_events: (pd.Series) of t_events. These are timestamps that will seed every triple barrier.
        These are the timestamps selected by the sampling procedures discussed in Chapter 2, Section 2.5.
        Eg: CUSUM Filter
    :param pt_sl: (2 element array) Element 0, indicates the profit taking level; Element 1 is stop loss level.
        A non-negative float that sets the width of the two barriers. A 0 value means that the respective
        horizontal barrier (profit taking and/or stop loss) will be disabled.
    :param target: (pd.Series) of values that are used (in conjunction with pt_sl) to determine the width
        of the barrier. In this program this is daily volatility series.
    :param min_ret: (float) The minimum target return required for running a triple barrier search.
    :param num_threads: (int) The number of threads concurrently used by the function.
    :param vertical_barrier_times: (pd.Series) A pandas series with the timestamps of the vertical barriers.
        We pass a False when we want to disable vertical barriers.
    :param side_prediction: (pd.Series) Side of the bet (long/short) as decided by the primary model
    :param verbose: (bool) Flag to report progress on asynch jobs
    :return: (pd.DataFrame) Events
            -events.index is event's starttime
            -events['t1'] is event's endtime
            -events['trgt'] is event's target
            -events['side'] (optional) implies the algo's position side
            -events['pt'] is profit taking multiple
            -events['sl']  is stop loss multiple
    """

    # 1) Get target
    target = target.reindex(t_events)
    target = target[target > min_ret]  # min_ret
    target = target.dropna()  # Drop NaN values

    # 2) Get vertical barrier (max holding period)
    if vertical_barrier_times is False:
        vertical_barrier_times = pd.Series(pd.NaT, index=t_events, dtype=t_events.dtype)

    # 3) Form events object, apply stop loss on vertical barrier
    if side_prediction is None:
        side_ = pd.Series(1.0, index=target.index)
        pt_sl_ = [pt_sl[0], pt_sl[0]]
    else:
        side_ = side_prediction.reindex(target.index)  # Subset side_prediction on target index.
        pt_sl_ = pt_sl[:2]

    # Create a new df with [v_barrier, target, side] and drop rows that are NA in target
    events = pd.concat({'t1': vertical_barrier_times, 'trgt': target, 'side': side_}, axis=1)
    events = events.dropna(subset=['trgt'])

    # Apply Triple Barrier
    first_touch_dates = mp_pandas_obj(func=apply_pt_sl_on_t1,
                                      pd_obj=('molecule', events.index),
                                      num_threads=1,  # Force single thread for now
                                      close=close,
                                      events=events,
                                      pt_sl=pt_sl_)

    for ind in events.index:
        events.at[ind, 't1'] = first_touch_dates.loc[ind, :].dropna().min()

    if side_prediction is None:
        events = events.drop('side', axis=1)

    # Add profit taking and stop loss multiples for vertical barrier calculations
    events['pt'] = pt_sl[0]
    events['sl'] = pt_sl[1]

    return events


# Snippet 3.9, pg 55, Question 3.3
def barrier_touched(out_df, events):
    """
    Advances in Financial Machine Learning, Snippet 3.9, page 55, Question 3.3.

    Adjust the getBins function (Snippet 3.7) to return a 0 whenever the vertical barrier is the one touched first.

    Top horizontal barrier: 1
    Bottom horizontal barrier: -1
    Vertical barrier: 0

    :param out_df: (pd.DataFrame) Returns and target
    :param events: (pd.DataFrame) The original events data frame. Contains the pt sl multiples needed here.
    :return: (pd.DataFrame) Returns, target, and labels
    """
    # Pre-allocate the store array
    store = np.zeros(len(out_df))
    
    for i, (date_time, values) in enumerate(out_df.iterrows()):
        ret = values['ret']
        target = values['trgt']

        pt_level_reached = ret > np.log(1 + target) * events.loc[date_time, 'pt']
        sl_level_reached = ret < -np.log(1 + target) * events.loc[date_time, 'sl']

        if ret > 0.0 and pt_level_reached:
            # Top barrier reached
            store[i] = 1
        elif ret < 0.0 and sl_level_reached:
            # Bottom barrier reached
            store[i] = -1
        # else: store[i] remains 0 for vertical barrier reached

    # Save to 'bin' column and return
    out_df['bin'] = store
    return out_df


# Snippet 3.4 -> 3.7, page 51, Labeling for Side & Size with Meta Labels
def get_bins(triple_barrier_events, close):
    """
    Advances in Financial Machine Learning, Snippet 3.7, page 51.

    Labeling for Side & Size with Meta Labels

    Compute event's outcome (including side information, if provided).
    events is a DataFrame where:

    Now the possible values for labels in out['bin'] are {0,1}, as opposed to whether to take the bet or pass,
    a purely binary prediction. When the predicted label the previous feasible values {−1,0,1}.
    The ML algorithm will be trained to decide is 1, we can use the probability of this secondary prediction
    to derive the size of the bet, where the side (sign) of the position has been set by the primary model.

    :param triple_barrier_events: (pd.DataFrame)
                -events.index is event's starttime
                -events['t1'] is event's endtime
                -events['trgt'] is event's target
                -events['side'] (optional) implies the algo's position side
                Case 1: ('side' not in events): bin in (-1,1) <-label by price action
                Case 2: ('side' in events): bin in (0,1) <-label by pnl (meta-labeling)
    :param close: (pd.Series) Close prices
    :return: (pd.DataFrame) Meta-labeled events
    """

    # 1) Align prices with their respective events
    events_ = triple_barrier_events.dropna(subset=['t1'])
    all_dates = events_.index.union(other=events_['t1'].array).drop_duplicates()
    prices = close.reindex(all_dates, method='bfill')

    # 2) Create out DataFrame
    out_df = pd.DataFrame(index=events_.index)
    # Need to take the log returns, else your results will be skewed for short positions
    out_df['ret'] = np.log(prices.loc[events_['t1'].array].array) - np.log(prices.loc[events_.index])
    out_df['trgt'] = events_['trgt']

    # Meta labeling: Events that were correct will have pos returns
    if 'side' in events_:
        out_df['ret'] = out_df['ret'] * events_['side']  # meta-labeling

    # Added code: label 0 when vertical barrier reached
    out_df = barrier_touched(out_df, triple_barrier_events)

    # Meta labeling: label incorrect events with a 0
    if 'side' in events_:
        out_df.loc[out_df['ret'] <= 0, 'bin'] = 0

    # Transform the log returns back to normal returns.
    out_df['ret'] = np.exp(out_df['ret']) - 1

    # Add the side to the output. This is useful for when a meta label model must be fit
    tb_cols = triple_barrier_events.columns
    if 'side' in tb_cols:
        out_df['side'] = triple_barrier_events['side']

    return out_df


# Snippet 3.8 page 54
def drop_labels(events, min_pct=.05):
    """
    Advances in Financial Machine Learning, Snippet 3.8 page 54.

    This function recursively eliminates rare observations.

    :param events: (dp.DataFrame) Events.
    :param min_pct: (float) A fraction used to decide if the observation occurs less than that fraction.
    :return: (pd.DataFrame) Events.
    """
    # Apply weights, drop labels with insufficient examples
    while True:
        df0 = events['bin'].value_counts(normalize=True)

        if df0.min() > min_pct or df0.shape[0] < 3:
            break

        print('dropped label: ', df0.idxmin(), df0.min())
        events = events[events['bin'] != df0.idxmin()]

    return events



# pylint: disable=missing-module-docstring
# pylint: disable=too-many-instance-attributes
import sys
import numpy as np
from famlafl.online_portfolio_selection.base import OLPS
from famlafl.online_portfolio_selection.crp import CRP


class UP(OLPS):
    """
    This class implements the Universal Portfolio strategy. It is reproduced with
    modification from the following paper:
    `Cover, T.M. (1991), Universal Portfolios. Mathematical Finance, 1: 1-29.
    <http://www-isl.stanford.edu/~cover/papers/portfolios_side_info.pdf>`_

    Universal Portfolio acts as a fund of funds, generating a number of experts with unique
    strategies. Cover's original universal portfolio integrates over the total simplex domain,
    but because it is not possible for us to calculate all possibilties, we generate a random
    distribution of points.

    The allocation method to each experts can be changed. If no
    allocation method is given, Universal Portfolio will not rebalance among the experts. Other
    allocation methods include uniform allocation among experts and top-k experts, which allocate
    capital based on the top-k performing experts until the last period.
    """
    def __init__(self, number_of_experts, weighted='hist_performance', k=1):
        """
        Initializes Universal Portfolio with the given number of experts, method of capital
        allocation to each experts, and k-value for Top-K experts.

        :param number_of_experts: (int) Number of total experts.
        :param weighted: (str) Capital allocation method. 'hist_performance': Historical Performance,
                               'uniform': Uniform Weights, 'top-k': Top-K experts.
        :param k: (int) Number of experts to choose your portfolio. Only necessary if weighted is
                        'top-k'. Typically lower values of k are more optimal for higher returns.
        """
        self.experts = []  # (list) Array to store all experts
        self.number_of_experts = number_of_experts  # (int) Set the number of experts.
        self.expert_params = None  # (np.array) Each expert's parameters.
        self.expert_portfolio_returns = None  # (np.array) All experts' portfolio returns over time.
        self.expert_all_weights = None  # (np.array) Each experts' weights over time.
        self.expert_weights = None  # (np.array) Each experts' final portfolio weights
        self.weights_on_experts = None  # (np.array) Capital allocation on each experts.
        self.weighted = weighted  # (np.array) Weights allocated to each experts.
        self.k = k  # (int) Number of top-k experts.
        super(UP, self).__init__()

    def _initialize(self, asset_prices, weights, resample_by):
        """
        Initializes the important variables for the object.

        :param asset_prices: (pd.DataFrame) Historical asset prices.
        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :param resample_by: (str) Specifies how to resample the prices. 'D' for Day, 'W' for Week,
                                 'M' for Month. The inputs are based on pandas' resample method.
        """
        # Initialize the same variables as OLPS.
        super(UP, self)._initialize(asset_prices, weights, resample_by)

        # Generate all experts.
        self._generate_experts()

        # Set experts' portfolio returns.
        self.expert_portfolio_returns = np.zeros((self.length_of_time, self.number_of_experts))
        # Set all experts' weights
        self.expert_all_weights = np.zeros((self.number_of_experts, self.length_of_time,
                                            self.number_of_assets))
        # Set all experts' predicted weights.
        self.expert_weights = np.zeros((self.number_of_experts, self.number_of_assets))

    def _run(self, weights, verbose):
        """
        Runs the algorithm by iterating through the given data.

        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :param verbose: (bool) Prints progress bar if true.
        """
        # Run allocate on all the experts.
        for exp in range(self.number_of_experts):
            # Allocate to each experts.
            self.experts[exp].allocate(self.asset_prices)
            # Stack the weights.
            self.expert_all_weights[exp] = self.experts[exp].all_weights
            # Stack the portfolio returns.
            self.expert_portfolio_returns[:, [exp]] = self.experts[exp].portfolio_return
            # Stack predicted weights.
            self.expert_weights[exp] = self.experts[exp].weights
            # Print progress bar.
            if verbose:
                self._print_progress(exp + 1, prefix='Progress:', suffix='Complete')
        # Calculate capital allocation on each experts.
        self._calculate_weights_on_experts()
        # Uniform weight distribution for wealth between managers.
        self._calculate_all_weights()

    def _calculate_weights_on_experts(self):
        """
        Calculates the weight allocation on each experts.
        'hist_performance': Historical Performance.
        'uniform': Uniform Weights.
        'top-k': Top-K experts.
        """
        # If capital allocation is based on historical performances.
        if self.weighted == 'hist_performance':
            # Calculate each expert's cumulative return ratio for each time period.
            expert_returns_ratio = np.apply_along_axis(lambda x: x/np.sum(x), 1,
                                                       self.expert_portfolio_returns)
            # Initial weights are evenly distributed among all experts.
            expert_returns_ratio = np.vstack((self._uniform_experts(), expert_returns_ratio))
            self.weights_on_experts = expert_returns_ratio

        # If capital allocation is based on uniform weights.
        elif self.weighted == 'uniform':
            # Equal allocation.
            uniform_ratio = np.ones(
                self.expert_portfolio_returns.shape) / self.number_of_experts
            # Initial weights are evenly distributed among all experts.
            uniform_ratio = np.vstack((self._uniform_experts(), uniform_ratio))
            self.weights_on_experts = uniform_ratio

        # If capital allocation is based on top-K experts.
        elif self.weighted == 'top-k':
            # Only the top k experts get 1/k of the wealth.
            # `<https://stackoverflow.com/questions/6910641/how-do-i-get-indices-of-n-maximum-values-in-a-numpy-array>`_
            # Get the indices of top k experts for each time.
            top_k = np.apply_along_axis(lambda x: np.argpartition(x, -self.k)[-self.k:], 1,
                                        self.expert_portfolio_returns)
            # Create a wealth distribution matrix.
            top_k_distribution = np.zeros(self.expert_portfolio_returns.shape)
            # For each week set the multiplier for each expert.
            # Each row represents the week's allocation to the k experts.

            for time in range(top_k.shape[0]):
                top_k_distribution[time][top_k[time]] = 1 / self.k
            # Initial weights are evenly distributed among all experts.
            top_k_distribution = np.vstack((self._uniform_experts(), top_k_distribution))
            self.weights_on_experts = top_k_distribution
        else:
            raise ValueError("Please put in 'hist_performance' for Historical Performance, "
                             "'uniform' for Uniform Distribution, or 'top-k' for top-K experts.")

    def recalculate_k(self, k):
        """
        Calculates the existing strategy with a different k value. The user does not have to
        rerun the entire strategy, but can simply recalculate with another k parameter.

        :param k: (int) Number of new top-k experts.
        """
        # Check that k value is an integer.
        if not isinstance(k, int):
            raise ValueError("K value must be an integer.")

        # Check that k value is at least 1.
        if k < 1:
            raise ValueError("K value must be greater than or equal to 1.")

        # Check that k value is less than window * rho.
        if k > self.number_of_experts:
            raise ValueError("K must be less than or equal to window * rho.")

        self.k = k

        # Calculate capital allocation on each experts.
        self._calculate_weights_on_experts()

        # Uniform weight distribution for wealth between managers.
        self._calculate_all_weights()

        # Round weights and drop values that are less than the given threshold.
        self._round_weights(threshold=1e-6)

        # Calculate portfolio returns based on weights calculated from the run method.
        self._calculate_portfolio_returns(self.all_weights, self.relative_return)

        # Convert everything to dataframe to make the information presentable.
        self._conversion()

    def _uniform_experts(self):
        """
        Returns a uniform weight of experts.

        :return: (np.array) Uniform weights (1/n, 1/n, 1/n ...).
        """
        # Divide by number of assets after creating numpy arrays of one.
        uni_weight = np.ones(self.number_of_experts) / self.number_of_experts
        return uni_weight

    def _calculate_all_weights(self):
        # pylint: disable=unsubscriptable-object
        """
        Calculate portfolio's overall weights and final predicted weights with information from
        each expert's weights.
        """

        # Calculate the product of the distribution matrix with the 3d experts x all weights matrix.
        # `<https://stackoverflow.com/questions/58588378/how-to-matrix-multiply-a-2d-numpy-array-with-a-3d-array-to-give-a-3d-array>`_
        d_shape = self.weights_on_experts[:-1].shape[:1] + self.expert_all_weights.shape[1:]
        reshaped_all_weights = self.expert_all_weights.reshape(self.expert_all_weights.shape[0], -1)
        weight_change = np.dot(self.weights_on_experts[:-1], reshaped_all_weights).reshape(d_shape)
        # We are looking at the diagonal cross section of the multiplication.
        self.all_weights = np.diagonal(weight_change, axis1=0, axis2=1).T

        # Calculate final predicted weights.
        self.weights = np.dot(self.weights_on_experts[[-1]], self.expert_weights)

    def _generate_experts(self):
        """
        Generate experts with the specified parameter.
        """
        # Generate randomized weights within the simplex.
        self._generate_simplex(self.number_of_experts, self.number_of_assets)
        # Universal Portfolio looks at different CRP weights.
        for exp in range(self.number_of_experts):
            self.experts.append(CRP(weight=self.expert_params[exp]))

    def _generate_simplex(self, number_of_experts, number_of_assets):
        """
        Generate uniform points on a simplex domain.
        `<https://cs.stackexchange.com/questions/3227/uniform-sampling-from-a-simplex>`_

        :param number_of_experts: (int) Number of experts.
        :param number_of_assets: (int) Number of assets.
        """
        # Create a randomized array with dimensions of (number_of_experts, number_of_assets - 1).
        simplex = np.sort(np.random.random((number_of_experts, number_of_assets - 1)))

        # Stack a column of zeros on the left, and stack a column of ones on the right.
        # Calculate the difference of each interval.
        # Each interval equates to a uniform sampling of the simplex domain.
        simplex = np.diff(np.hstack([np.zeros((number_of_experts, 1)), simplex,
                                     np.ones((number_of_experts, 1))]))
        # Set the parameters as the randomly generated simplex weights.
        self.expert_params = simplex

    def _print_progress(self, iteration, prefix='', suffix='', decimals=1, bar_length=50):
        # pylint: disable=expression-not-assigned
        """
        Calls in a loop to create a terminal progress bar.
        `<https://gist.github.com/aubricus/f91fb55dc6ba5557fbab06119420dd6a>`_

        :param iteration: (int) Current iteration.
        :param prefix: (str) Prefix string.
        :param suffix: (str) Suffix string.
        :param decimals: (int) Positive number of decimals in percent completed.
        :param bar_length: (int) Character length of the bar.
        """
        str_format = "{0:." + str(decimals) + "f}"
        # Calculate the percent completed.
        percents = str_format.format(100 * (iteration / float(self.number_of_experts)))
        # Calculate the length of bar.
        filled_length = int(round(bar_length * iteration / float(self.number_of_experts)))
        # Fill the bar.
        block = '█' * filled_length + '-' * (bar_length - filled_length)
        # Print new line.
        sys.stdout.write('\r%s |%s| %s%s %s' % (prefix, block, percents, '%', suffix)),

        if iteration == self.number_of_experts:
            sys.stdout.write('\n')
        sys.stdout.flush()



# pylint: disable=missing-module-docstring
from famlafl.online_portfolio_selection.base import OLPS


class CRP(OLPS):
    """
    This class implements the Buy and Hold strategy. It is reproduced with modification from
    the following paper:
    `Li, B., Hoi, S. C.H., 2012. OnLine Portfolio Selection: A Survey. ACM Comput.
    Surv. V, N, Article A (December 2012), 33 pages. <https://arxiv.org/abs/1212.2129>`_

    Constant Rebalanced Portfolio rebalances to a given weight each time period.
    """
    def __init__(self, weight=None):
        """
        Sets the recurring weights for the Constant Rebalanced Portfolio. If weight is given,
        this will override any given weights inputted by the user through ``allocate``.

        :param weight: (list/np.array/pd.DataFrame) Initial weight set by the user.
        """
        super(CRP, self).__init__()
        self.weight = weight

    def _first_weight(self, weights):
        """
        Sets first weight for Constant Rebalanced Portfolio

        :param weights: (list/np.array/pd.DataFrame) initial weights set by the user.
        :return: (np.array) First portfolio weight.
        """
        # Initialize with given weights.
        if self.weight is not None:
            return self.weight
        # If no weights are given, automatically set weights to uniform weights.
        if weights is None:
            weights = self._uniform_weight()
        return weights



"""
Test Trend Scanning labels
"""

import os
import unittest
import pandas as pd

from famlafl.labeling import trend_scanning_labels


class TestTrendScanningLabels(unittest.TestCase):
    """
    Test trend-scanning labels
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """
        project_path = os.path.dirname(__file__)

        # Data set used for trend scanning labels
        self.eem_close = pd.read_csv(project_path + '/test_data/stock_prices.csv', index_col=0, parse_dates=[0])
        # In 2008, EEM had some clear trends
        self.eem_close = self.eem_close['EEM'].loc[pd.Timestamp(2008, 4, 1):pd.Timestamp(2008, 10, 1)]

    def test_trend_scanning_labels(self):
        """
        Test trend scanning labels
        """

        t_events = self.eem_close.index
        tr_scan_labels = trend_scanning_labels(self.eem_close, t_events, 20)

        self.assertEqual(tr_scan_labels.shape[0], len(t_events))  # we have label value for all t events

        # Before 2008/5/12 we had a strong positive trend
        self.assertTrue(
            set(tr_scan_labels.loc[pd.Timestamp(2008, 1, 1):pd.Timestamp(2008, 5, 9)].bin) == set([1]))

        self.assertEqual(tr_scan_labels.bin.value_counts()[-1], 70)  # Number of -1 labels check
        self.assertEqual(tr_scan_labels.bin.value_counts()[1], 40)

        # Approx. 20 latest labels are nan because we don't have data for look-forward period (20)
        self.assertEqual(tr_scan_labels.shape[0] - tr_scan_labels.dropna().shape[0], 19)

        tr_scan_labels.dropna(inplace=True)  # Drop na values

        # Index should be < t1
        self.assertTrue((tr_scan_labels.t1 > tr_scan_labels.index).all())

        for int_index, (ret_v, bin_v) in zip([0, 2, 10, 20, 50],
                                             [(0.05037, 1), (0.0350, 1), (0.07508, 1), (0.05219, 1), (0.02447, 1)]):
            self.assertAlmostEqual(tr_scan_labels.iloc[int_index]['ret'], ret_v, delta=1e-4)
            self.assertEqual(tr_scan_labels.iloc[int_index]['bin'], bin_v)

        tr_scan_labels_none = trend_scanning_labels(self.eem_close, t_events=None, look_forward_window=20)
        tr_scan_labels_none.dropna(inplace=True)

        self.assertTrue((tr_scan_labels == tr_scan_labels_none).all().all())



"""
Implementations of mutual information (I) and variation of information (VI) codependence measures from Cornell
lecture slides: https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3512994&download=yes
"""
import numpy as np
import scipy.stats as ss
from sklearn.metrics import mutual_info_score


# pylint: disable=invalid-name

def get_optimal_number_of_bins(num_obs: int, corr_coef: float = None) -> int:
    """
    Calculates optimal number of bins for discretization based on number of observations
    and correlation coefficient (univariate case).

    Algorithms used in this function were originally proposed in the works of Hacine-Gharbi et al. (2012)
    and Hacine-Gharbi and Ravier (2018). They are described in the Cornell lecture notes:
    https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3512994&download=yes (p.26)

    :param num_obs: (int) Number of observations.
    :param corr_coef: (float) Correlation coefficient, used to estimate the number of bins for univariate case.
    :return: (int) Optimal number of bins.
    """

    # Univariate case
    if corr_coef is None or abs(corr_coef - 1) <= 1e-4:
        z = (8 + 324 * num_obs + 12 * (36 * num_obs + 729 * num_obs ** 2) ** .5) ** (1 / 3.)
        bins = round(z / 6. + 2. / (3 * z) + 1. / 3)

    # Bivariate case
    else:
        bins = round(2 ** -.5 * (1 + (1 + 24 * num_obs / (1. - corr_coef ** 2)) ** .5) ** .5)
    return int(bins)


def get_mutual_info(x: np.array, y: np.array, n_bins: int = None, normalize: bool = False) -> float:
    """
    Returns mutual information (I) between two vectors.

    This function uses the discretization with the optimal bins algorithm proposed in the works of
    Hacine-Gharbi et al. (2012) and Hacine-Gharbi and Ravier (2018).

    Read Cornell lecture notes for more information about the mutual information:
    https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3512994&download=yes.

    :param x: (np.array) X vector.
    :param y: (np.array) Y vector.
    :param n_bins: (int) Number of bins for discretization, if None the optimal number will be calculated.
                         (None by default)
    :param normalize: (bool) Flag used to normalize the result to [0, 1]. (False by default)
    :return: (float) Mutual information score.
    """

    if n_bins is None:
        corr_coef = np.corrcoef(x, y)[0][1]
        n_bins = get_optimal_number_of_bins(x.shape[0], corr_coef=corr_coef)

    contingency = np.histogram2d(x, y, n_bins)[0]
    mutual_info = mutual_info_score(None, None, contingency=contingency)  # Mutual information
    if normalize is True:
        marginal_x = ss.entropy(np.histogram(x, n_bins)[0])  # Marginal for x
        marginal_y = ss.entropy(np.histogram(y, n_bins)[0])  # Marginal for y
        mutual_info /= min(marginal_x, marginal_y)
    return mutual_info


def variation_of_information_score(x: np.array, y: np.array, n_bins: int = None, normalize: bool = False) -> float:
    """
    Returns variantion of information (VI) between two vectors.

    This function uses the discretization using optimal bins algorithm proposed in the works of
    Hacine-Gharbi et al. (2012) and Hacine-Gharbi and Ravier (2018).

    Read Cornell lecture notes for more information about the variation of information:
    https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3512994&download=yes.

    :param x: (np.array) X vector.
    :param y: (np.array) Y vector.
    :param n_bins: (int) Number of bins for discretization, if None the optimal number will be calculated.
                         (None by default)
    :param normalize: (bool) True to normalize the result to [0, 1]. (False by default)
    :return: (float) Variation of information score.
    """

    if n_bins is None:
        corr_coef = np.corrcoef(x, y)[0][1]
        n_bins = get_optimal_number_of_bins(x.shape[0], corr_coef=corr_coef)

    contingency = np.histogram2d(x, y, n_bins)[0]
    mutual_info = mutual_info_score(None, None, contingency=contingency)  # Mutual information
    marginal_x = ss.entropy(np.histogram(x, n_bins)[0])  # Marginal for x
    marginal_y = ss.entropy(np.histogram(y, n_bins)[0])  # Marginal for y
    score = marginal_x + marginal_y - 2 * mutual_info  # Variation of information

    if normalize is True:
        joint_dist = marginal_x + marginal_y - mutual_info  # Joint distribution
        score /= joint_dist

    return score



"""
Initialize Benchmarks Module.
"""
from famlafl.online_portfolio_selection.bah import BAH
from famlafl.online_portfolio_selection.best_stock import BestStock
from famlafl.online_portfolio_selection.crp import CRP
from famlafl.online_portfolio_selection.bcrp import BCRP



# pylint: disable=no-self-use
# pylint: disable=unnecessary-comprehension
"""
Matrix Flag labeling method.
"""

import pandas as pd
import numpy as np


class MatrixFlagLabels:
    """
    The Matrix Flag labeling method is featured in the paper: Cervelló-Royo, R., Guijarro, F. and Michniuk, K., 2015.
    Stock market trading rule based on pattern recognition and technical analysis: Forecasting the DJIA index with
    intraday data.

    The method of applying a matrix template was first introduced, and explained in greater detail, in the paper:
    Leigh, W., Modani, N., Purvis, R. and Roberts, T., 2002. Stock market trading rule discovery using technical
    charting heuristics.

    Cervelló-Royo et al. expand on Leigh et al.'s work by proposing a new bull flag pattern which ameliorates some
    weaknesses in Leigh's original template. Additionally, he applies this bull flag labeling method to intraday
    candlestick data, rather than just closing prices.

    To find the total weight for a given day, the current price as well as the preceding window days number of prices is
    used. The data window is split into 10 buckets each containing a chronological tenth of the data window. Each point
    in 1 bucket is put into a decile corresponding to a position in a column based on percentile relative to the entire
    data window. Bottom 10% on lowest row, next 10% on second lowest row etc.
    The proportion of points in each decile is reported to finalize the column. The first tenth of the data is
    transformed to the leftmost column, the next tenth to the next column on the right and so on until finally a 10 by
    10 matrix is achieved. This matrix is then multiplied element-wise with the 10 by 10 template, and the sum of all
    columns is the total weight for the day. If desired, the user can specify a threshold to determine positive and
    negative classes. The value of the threshold depends on how strict of a classifier the user desires, and the
    allowable values based on the template matrix.
    """

    def __init__(self, prices, window, template_name=None):
        """
        :param prices: (pd.Series) Price data for one stock.
        :param window: (int) Length of preceding data window used when generating the fit matrix for one day.
        :param template_name: (str) Name of the an available template in the template library. Allowable names:
                            ``leigh_bear``, ``leigh_bull``, ``cervelloroyo_bear``, ``cervellororo_bull``.
        """
        assert (len(prices) >= 10), "Length of data must be at least 10."
        assert (window >= 10), "Window must be at least 10."
        assert (len(prices) >= window), "Window cannot be greater than length of data."
        assert isinstance(prices, pd.Series), "Data must be pd.Series."
        self.data = prices
        self.window = window
        self.template = pd.DataFrame()

        if template_name is not None:
            self._init_template(template_name)

    def _init_template(self, name):
        """
        :param name: (str) Name of the an available template in the template library. Allowable names: ``leigh_bear``,
        ``leigh_bull``, ``cervelloroyo_bear``, ``cervellororo_bull``.
        """
        leigh_bull = pd.DataFrame([[.5, 0, -1, -1, -1, -1, -1, -1, -1, 0],
                                   [1, 0.5, 0, -0.5, -1, -1, -1, -1, -0.5, 0],
                                   [1, 1, 0.5, 0, -0.5, -0.5, -0.5, -0.5, 0, 0.5],
                                   [0.5, 1, 1, 0.5, 0, -0.5, -0.5, -0.5, 0, 1],
                                   [0, 0.5, 1, 1, 0.5, 0, 0, 0, 0.5, 1],
                                   [0, 0, 0.5, 1, 1, 0.5, 0, 0, 1, 1],
                                   [-0.5, 0, 0, 0.5, 1, 1, 0.5, 0.5, 1, 1],
                                   [-0.5, -1, 0, 0, 0.5, 1, 1, 1, 1, 0],
                                   [-1, -1, -1, -0.5, 0, 0.5, 1, 1, 0, -2],
                                   [-1, -1, -1, -1, -1, 0, 0.5, 0.5, -2, -2.5]])

        leigh_bear = pd.DataFrame(np.flip(np.array(leigh_bull), axis=0))
        cervelloroyo_bull = pd.DataFrame([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                          [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                          [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                          [0, 0, 0, 0, -1, -1, -1, -1, -1, -1],
                                          [0, 0, 0, -1, -2, -2, -2, -2, -2, -2],
                                          [0, 0, -1, -3, -3, -3, -3, -3, -3, -3],
                                          [0, -1, -3, -5, -5, -5, -5, -5, -5, -5],
                                          [0, -1, -5, -5, -5, -5, -5, -5, -5, -5],
                                          [0, -1, -5, -5, -5, -5, -5, -5, -5, -5],
                                          [5, -1, -5, -5, -5, -5, -5, -5, -5, -5]])
        cervelloroyo_bear = pd.DataFrame(np.flip(np.array(cervelloroyo_bull), axis=0))

        if name == 'leigh_bull':
            self.set_template(leigh_bull)
        elif name == 'leigh_bear':
            self.set_template(leigh_bear)
        elif name == 'cervelloroyo_bear':
            self.set_template(cervelloroyo_bear)
        elif name == 'cervelloroyo_bull':
            self.set_template(cervelloroyo_bull)
        else:
            raise Exception("Invalid template name. Valid names are 'leigh_bull', 'leigh_bear', "
                            "'cervelloroyo_bear', 'cervelloroyo_bull'.")

    def set_template(self, template):
        """
        :param template: (pd.DataFrame) Template to override the default template. Must be a 10 by 10 pd.DataFrame.
                            NaN values not allowed, as they will not automatically be treated as zeros.
        """
        assert template.shape == (10, 10), "Template must be 10 by 10."
        assert not template.isnull().values.any(), "No NaN values allowed in template."
        self.template = template

    def _transform_data(self, row_num, window=30):
        """
        :param row_num: (int) Row number to use for the "current" data point to apply the window to. The data window
                        contains the row corresponding to row_num, as well as the (self.window-1) preceding rows.
        :param window: (int) The number of rows preceding the current one to use for window. Override with
                        self.window in most cases.
        :return: (pd.DataFrame) Transformed 10 by 10 matrix, in which each column corresponds to a chronological tenth
                    of the data window, and each row corresponds to a price decile relative to the entire data window.
                    The template matrix is then applied to this output matrix.
        """
        # The relevant data to create matrix for current day consists of the day's return and the window number of days
        # preceding.
        data_window = self.data[row_num - window: row_num]

        # Find values for cutoff percentiles (deciles) over the entire data window. percentile_cutoffs is a list of
        # cutoffs for the 10th, 20th, 30th etc. percentiles in the data window.
        percentiles = np.linspace(10, 100, num=10)
        percentile_cutoffs = np.percentile(data_window, percentiles)

        # Each value in bins corresponds to the value of the same index in data_window. If bin is 0, the corresponding
        # point in data_window is in the 0-10th percentile, if 1 then 10-20th percentile, ..., if 9 then 90-100th
        # percentile.
        bins = np.digitize(data_window, bins=percentile_cutoffs, right=True)

        # Dictionary to map values in data window to their percentile bin.
        value_to_bin = dict(zip(data_window, bins))

        # Split data window into 10 chronological sub-windows.
        data_split = np.array_split(data_window, 10)

        # Create the matrix, one column at a time from left to right. The leftmost column corresponds to the first
        # tenth of data, and the rightmost corresponds to the final tenths. The top element of each column corresponds
        # to the count in the tenth in the 90-100th percentile of data window, second from top is count in the tenth
        # in the 80-90th percentile and so on until the last element is the count in the 0-10th percentile.
        matrix = pd.DataFrame()
        col_num = 0  # Left most column (out of 10) to populate using the first tenth of data.
        for tenth in data_split:
            # Apply the dictionary to data to get the bins each data point belongs to.
            tenth_bins = np.vectorize(value_to_bin.get)(tenth)
            # We count the number of times each bin appears in each tenth of data, starting from the highest decile.
            column, _ = np.histogram(tenth_bins, bins=[i for i in range(11)])
            # Convert the count in the column to proportion.
            column = np.array(column) / len(tenth)
            matrix[col_num] = column[::-1]  # Reverse so that highest decile is on top of the column.
            col_num += 1

        return matrix

    def _apply_template_to_matrix(self, matrix, template):
        """
        :param matrix: (pd.DataFrame) Processed 10 by 10 matrix, where each column represents a chronological tenth
                        of the data, and each row represents a decile relative to the entire data window.
        :param template: (pd.DataFrame) Template to apply the processed matrix to.
        :return: (float) The total score for the day. Consists of the sum of sum of columns of the result from
                    multiplying the matrix element-wise with the template.
        """
        new_mat = matrix * template
        total_fit = new_mat.values.sum()
        return total_fit

    def apply_labeling_matrix(self, threshold=None):
        """
        :param threshold: (float) If None, labels will be returned numerically as the score for the day. If not None,
                        then labels are returned categorically, with the positive category for labels that are equal to
                        or exceed the threshold.
        :return: (pd.Series) Total scores for the data series on each eligible day (meaning for indices self.window and
                    onwards).
        """
        labels = []
        idx = self.data[self.window:len(self.data)].index

        # Apply the data transformation and template.
        for row_num in range(self.window, len(self.data)):
            weights_matrix = self._transform_data(row_num=row_num, window=self.window)
            label = self._apply_template_to_matrix(weights_matrix, self.template)
            labels.append(label)

        # If threshold is given, replace fit values by whether they exceed the threshold.
        if threshold is not None:
            labels = [1 if i >= threshold else -1 for i in labels]
        return pd.Series(data=labels, index=idx)



"""
An implementation of the Exact Fit of the first 3 Moments (EF3M) of finding the parameters that make up the mixture
of 2 Gaussian distributions. Based on the work by Lopez de Prado and Foreman (2014) "A mixture of two Gaussians
approach to mathematical portfolio oversight: The EF3M algorithm." Quantitative Finance, Vol. 14, No. 5, pp. 913-930.
"""

import sys
from multiprocessing import cpu_count, Pool
import numpy as np
import pandas as pd
from scipy.special import comb
from scipy.stats import gaussian_kde
from numba import njit, objmode, float64


class M2N:
    """
    M2N - A Mixture of 2 Normal distributions
    This class is used to contain parameters and equations for the EF3M algorithm, when fitting parameters to a mixture
    of 2 Gaussian distributions.

    :param moments: (list) The first five (1... 5) raw moments of the mixture distribution.
    :param epsilon: (float) Fitting tolerance
    :param factor: (float) Lambda factor from equations
    :param n_runs: (int) Number of times to execute 'singleLoop'
    :param variant: (int) The EF3M variant to execute, options are 1: EF3M using first 4 moments, 2: EF3M using
     first 5 moments
    :param max_iter: (int) Maximum number of iterations to perform in the 'fit' method
    :param num_workers: (int) Number of CPU cores to use for multiprocessing execution. Default is -1 which sets
     num_workers to all cores.

    """
    def __init__(self, moments, epsilon=10**-5, factor=5, n_runs=1, variant=1, max_iter=100_000, num_workers=-1):
        """
        Constructor

        :param moments: (list) The first five (1... 5) raw moments of the mixture distribution.
        :param epsilon: (float) Fitting tolerance
        :param factor: (float) Lambda factor from equations
        :param n_runs: (int) Number of times to execute 'singleLoop'
        :param variant: (int) The EF3M variant to execute, options are 1: EF3M using first 4 moments, 2: EF3M using
         first 5 moments
        :param max_iter: (int) Maximum number of iterations to perform in the 'fit' method
        :param num_workers: (int) Number of CPU cores to use for multiprocessing execution. Default is -1 which sets
         num_workers to all cores.

        The parameters of the mixture are defined by a list, where:
            parameters = [mu_1, mu_2, sigma_1, sigma_2, p_1]
        """
        # Set fitting parameters in constructor.
        self.epsilon = epsilon
        self. factor = factor
        self.n_runs = n_runs
        self.variant = variant
        self.max_iter = max_iter
        self.num_workers = num_workers
        # Set moments to fit and initialize lists and errors.
        self.moments = moments
        self.new_moments = [0 for _ in range(5)]  # Initialize the new moment list to zeroes.
        self.parameters = [0 for _ in range(5)]  # Initialize the parameter list to zeroes.
        self.error = sum([moments[i]**2 for i in range(len(moments))])

    def fit(self, mu_2):
        """
        Fits and the parameters that describe the mixture of the 2 Normal distributions for a given set of initial
        parameter guesses.

        :param mu_2: (float) An initial estimate for the mean of the second distribution.
        """
        p_1 = np.random.uniform(0, 1)
        num_iter = 0
        while True:
            num_iter += 1
            if self.variant == 1:
                parameters_new = self.iter_4(mu_2, p_1)  # First variant, using the first 4 moments.
            elif self.variant == 2:
                parameters_new = self.iter_5(mu_2, p_1)  # Second variant, using all 5 moments.
            else:
                raise ValueError("Value of argument 'variant' must be either 1 or 2.")

            if not parameters_new:
                # An empty list returned means an invalid value was found in iter_4 or iter_5.
                return None

            parameters = parameters_new.copy()
            self.get_moments(parameters)
            error = sum([(self.moments[i] - self.new_moments[i])**2 for i in range(len(self.new_moments))])
            if error < self.error:
                # Update with new best parameters, error.
                self.parameters = parameters
                self.error = error

            if abs(p_1 - parameters[4]) < self.epsilon:
                # Stopping condition.
                break

            if num_iter > self.max_iter:
                # Stops calculation if algorithm reaches the set maximum number of iterations.
                return None

            p_1 = parameters[4]
            mu_2 = parameters[1]  # Update for the 5th moments convergence.

        self.parameters = parameters
        return None

    def get_moments(self, parameters, return_result=False):
        """
        Calculates and returns the first five (1...5) raw moments corresponding to the newly estimated parameters.

        :param parameters: (list) List of parameters if the specific order [mu_1, mu_2, sigma_1, sigma_2, p_1]
        :param return_result: (bool) If True, method returns a result instead of setting the 'self.new_moments'
         attribute.
        :return: (list) List of the first five moments
        """
        u_1, u_2, s_1, s_2, p_1 = parameters  # Expanded mixture parameters to individual variables for clarity.
        p_2 = 1 - p_1  # Explicitly state p_2 for symmetry.
        m_1 = p_1 * u_1 + p_2 * u_2  # Eq. (6)
        m_2 = p_1 * (s_1**2 + u_1**2) + p_2 * (s_2**2 + u_2**2)  # Eq. (7)
        m_3 = p_1 * (3 * s_1**2 * u_1 + u_1**3) + p_2 * (3 * s_2**2 * u_2 + u_2**3)  # Eq. (8)
        # Eq. (9)
        m_4 = p_1 * (3 * s_1**4 + 6 * s_1**2 * u_1**2 + u_1**4) + p_2 * (3 * s_2**4 + 6 * s_2**2 * u_2**2 + u_2**4)
        # Eq (10)
        m_5 = p_1 * (15 * s_1**4 * u_1 + 10 * s_1**2 * u_1**3 + u_1**5) + p_2 *\
            (15 * s_2**4 * u_2 + 10 * s_2**2 * u_2**3 + u_2**5)

        if return_result:
            return [m_1, m_2, m_3, m_4, m_5]

        self.new_moments = [m_1, m_2, m_3, m_4, m_5]
        return None

    def iter_4(self, mu_2, p_1):
        """
        Evaluation of the set of equations that make up variant #1 of the EF3M algorithm (fitting using the first
        four moments).

        :param mu_2: (float) Initial parameter value for mu_2
        :param p_1: (float) Probability defining the mixture; p_1, 1 - p_1
        :return: (list) List of estimated parameter if no invalid values are encountered (e.g. complex values,
         divide-by-zero), otherwise an empty list is returned.
        """
        # Expand list of moments to individual variables for clarity.
        m_1, m_2, m_3, m_4 = self.moments[0:4]

        # Check to see if every value made it through.
        param_list = iter_4_jit(mu_2, p_1, m_1, m_2, m_3, m_4)
        param_list = param_list.tolist()

        if len(param_list) < 5:
            return []

        return param_list

    def iter_5(self, mu_2, p_1):
        """
        Evaluation of the set of equations that make up variant #2 of the EF3M algorithm (fitting using the first five
        moments).

        :param mu_2: (float) Initial parameter value for mu_2
        :param p_1: (float) Probability defining the mixture; p_1, 1-p_1
        :return: (list) List of estimated parameter if no invalid values are encountered (e.g. complex values,
         divide-by-zero), otherwise an empty list is returned.
        """

        # Expand list of moments to individual variables for clarity.
        (m_1, m_2, m_3, m_4, m_5,) = self.moments

        # Call numba decorated function to do the actual calculations
        param_list = iter_5_jit(mu_2, p_1, m_1, m_2, m_3, m_4, m_5)

        param_list = param_list.tolist()

        if len(param_list) < 5:
            return []

        return param_list

    def single_fit_loop(self, epsilon=0):
        """
        A single scan through the list of mu_2 values, cataloging the successful fittings in a DataFrame.

        :param epsilon: (float) Fitting tolerance.
        :return: (pd.DataFrame) Fitted parameters and error
        """
        # Reset parameters and error for each single_fit_loop.
        self.epsilon = epsilon if epsilon != 0 else self.epsilon
        self.parameters = [0 for _ in range(5)]  # Initialize the parameter list.
        self.error = sum([self.moments[i]**2 for i in range(len(self.moments))])

        std_dev = centered_moment(self.moments, 2)**0.5
        mu_2 = [float(i) * self.epsilon * self.factor * std_dev + self.moments[0] for i in range(1, int(1/self.epsilon))]
        err_min = self.error

        d_results = {}
        for mu_2_i in mu_2:
            self.fit(mu_2=mu_2_i)

            if self.error < err_min:
                err_min = self.error
                d_results['mu_1'], d_results['mu_2'], d_results['sigma_1'], d_results['sigma_2'], d_results['p_1'] = \
                    [[p] for p in self.parameters]
                d_results['error'] = [err_min]

        return pd.DataFrame.from_dict(d_results)

    def mp_fit(self):
        """
        Parallelized implementation of the 'single_fit_loop' method. Makes use of dask.delayed to execute multiple
        calls of 'single_fit_loop' in parallel.

        :return: (pd.DataFrame) Fitted parameters and error
        """
        num_workers = self.num_workers if self.num_workers > 0 else cpu_count()
        pool = Pool(num_workers)

        output_list = pool.imap_unordered(self.single_fit_loop, [self.epsilon for i in range(self.n_runs)])
        df_list = []

        # Process asynchronous output, report progress and progress bar.
        max_prog_bar_len = 25
        for i, out_i in enumerate(output_list, 1):
            df_list.append(out_i)
            num_fill = int((i/self.n_runs) * max_prog_bar_len)
            prog_bar_string = '|' + num_fill*'#' + (max_prog_bar_len-num_fill)*' ' + '|'
            sys.stderr.write(f'\r{prog_bar_string} Completed {i} of {self.n_runs} fitting rounds.')
        # Close and clean up pool.
        pool.close()
        pool.join()
        # Concatenate and return results of fitting.
        df_out = pd.concat(df_list)
        return df_out


# === Helper functions, outside the M2N class. === #
def centered_moment(moments, order):
    """
    Compute a single moment of a specific order about the mean (centered) given moments about the origin (raw).

    :param moments: (list) First 'order' raw moments
    :param order: (int) The order of the moment to calculate
    :return: (float) The central moment of specified order.
    """
    moment_c = 0  # First centered moment is always zero.
    for j in range(order + 1):
        combin = int(comb(order, j))
        if j == order:
            a_1 = 1
        else:
            a_1 = moments[order - j - 1]
        moment_c += (-1)**j * combin * moments[0]**j * a_1

    return moment_c


def raw_moment(central_moments, dist_mean):
    """
    Calculates a list of raw moments given a list of central moments.

    :param central_moments: (list) The first n (1...n) central moments as a list.
    :param dist_mean: (float) The mean of the distribution.
    :return: (list) The first n+1 (0...n) raw moments.
    """
    raw_moments = [dist_mean]
    central_moments = [1] + central_moments  # Add the zeroth moment to the front of the list, just helps with indexing.
    for n_i in range(2, len(central_moments)):
        moment_n_parts = []
        for k in range(n_i + 1):
            sum_part = comb(n_i, k) * central_moments[k] * dist_mean**(n_i - k)
            moment_n_parts.append(sum_part)
        moment_n = sum(moment_n_parts)
        raw_moments.append(moment_n)
    return raw_moments


def most_likely_parameters(data, ignore_columns='error', res=10_000):
    """
    Determines the most likely parameter estimate using a KDE from the DataFrame of the results of the fit from the
    M2N object.

    :param data: (pandas.DataFrame) Contains parameter estimates from all runs.
    :param ignore_columns: (string, list) Column or columns to exclude from analysis.
    :param res: (int) Resolution of the kernel density estimate.
    :return: (dict) Labels and most likely estimates for parameters.
    """
    df_results = data.copy()
    if isinstance(ignore_columns, str):
        ignore_columns = [ignore_columns]

    columns = [c for c in df_results.columns if c not in ignore_columns]
    d_results = {}
    for col in columns:
        x_range = np.linspace(df_results[col].min(), df_results[col].max(), num=res)
        kde = gaussian_kde(df_results[col].to_numpy())
        y_kde = kde.evaluate(x_range)
        top_value = round(x_range[np.argmax(y_kde)], 5)
        d_results[col] = top_value

    return d_results


@njit
def iter_4_jit(mu_2: float64, p_1: float64, m_1: float64, m_2: float64, m_3: float64, m_4: float64) -> np.ndarray:  # pragma: no cover
    """
    "Numbarized" evaluation of the set of equations that make up variant #1 of the EF3M algorithm (fitting using the
    first four moments).

    :param mu_2: (float) Initial parameter value for mu_2
    :param p_1: (float) Probability defining the mixture; p_1, 1 - p_1
    :param m_1, m_2, m_3, m_4: (float) The first four (1... 4) raw moments of the mixture distribution.
    :return: (list) List of estimated parameter if no invalid values are encountered (e.g. complex values,
        divide-by-zero), otherwise an empty list is returned.
    """
    param_list = np.empty(0, dtype=float64)

    # Using a while-loop here to be able to use 'break' functionality.
    # We need to stop the calculation at any given step to avoid throwing warnings or errors,
    # and be in control of our return values. I'm open to other suggestions, but multiple return statements isn't
    # one of them.
    while True:
        # Calculate mu_1, Equation (22).
        mu_1 = (m_1 - (1 - p_1) * mu_2) / p_1

        # Calculate sigma_2, Equation (24)
        if (3 * (1 - p_1) * (mu_2 - mu_1)) == 0:
            # Validity check 1: Check for divide-by-zero.
            break

        sigma_2_squared = (
            m_3
            + 2 * p_1 * mu_1 ** 3
            + (p_1 - 1) * mu_2 ** 3
            - 3 * mu_1 * (m_2 + mu_2 ** 2 * (p_1 - 1))
        ) / (3 * (1 - p_1) * (mu_2 - mu_1))
        if sigma_2_squared < 0:
            # Validity check 2: Prevent potential complex values.
            break

        sigma_2 = sigma_2_squared ** 0.5
        # Calculate sigma_1, Equation (23)
        sigma_1_squared = (
            (m_2 - sigma_2 ** 2 - mu_2 ** 2) / p_1
            + sigma_2 ** 2
            + mu_2 ** 2
            - mu_1 ** 2
        )

        if sigma_1_squared < 0:
            # Validity check 3: Prevent potential complex values.
            break
        sigma_1 = sigma_1_squared ** 0.5

        # Adjust guess for p_1, Equation (25)
        p_1_deno = (
            3 * (sigma_1 ** 4 - sigma_2 ** 4)
            + 6 * (sigma_1 ** 2 * mu_1 ** 2 - sigma_2 ** 2 * mu_2 ** 2)
            + mu_1 ** 4
            - mu_2 ** 4
        )
        if p_1_deno == 0:
            # Validity check 5: Break if about to divide by zero.
            break

        p_1 = (
            m_4 - 3 * sigma_2 ** 4 - 6 * sigma_2 ** 2 * mu_2 ** 2 - mu_2 ** 4
        ) / p_1_deno
        if (p_1 < 0) or (p_1 > 1):
            # Validity check 6: The probability must be between zero and one.
            break

        # Add all new parameter estimates to the return list if no break has occurred before now.
        param_list = np.array([mu_1, mu_2, sigma_1, sigma_2, p_1], dtype=float64)

        # We only want this to execute once at most, so call a final break if one hasn't been called yet.
        break

    return param_list


@njit
def iter_5_jit(mu_2: float64, p_1: float64, m_1: float64, m_2: float64, m_3: float64, m_4: float64, m_5: float64) -> np.ndarray:  # pragma: no cover
    """
    "Numbarized" evaluation of the set of equations that make up variant #2 of the EF3M algorithm (fitting using the
     first five moments).

    :param mu_2: (float) Initial parameter value for mu_2
    :param p_1: (float) Probability defining the mixture; p_1, 1-p_1
    :param m_1, m_2, m_3, m_4, m_5: (float) The first five (1... 5) raw moments of the mixture distribution.
    :return: (list) List of estimated parameter if no invalid values are encountered (e.g. complex values,
        divide-by-zero), otherwise an empty list is returned.
    """
    param_list = np.empty(0, dtype=float64)

    # Using a while-loop here to be able to use 'break' functionality.
    # We need to stop the calculation at any given step to avoid throwing warnings or errors, and be in control
    # of our return values. I'm open to other suggestions, but multiple return statements isn't one of them.
    while True:
        # Calculate mu_1, Equation (22).
        mu_1 = (m_1 - (1 - p_1) * mu_2) / p_1
        if (3 * (1 - p_1) * (mu_2 - mu_1)) == 0:
            # Validity check 1: check for divide-by-zero.
            break

        # Calculate sigma_2, Equation (24).
        sigma_2_squared = (m_3 + 2 * p_1 * mu_1 ** 3 + (p_1 - 1) * mu_2 ** 3 - 3 * mu_1 * (m_2 + mu_2 ** 2 * (p_1 - 1))
                           ) / (3 * (1 - p_1) * (mu_2 - mu_1))

        if sigma_2_squared < 0:
            # Validity check 2: check for upcoming complex numbers.
            break

        sigma_2 = sigma_2_squared ** 0.5

        # Calculate sigma_1, Equation (23).
        sigma_1_squared = ((m_2 - sigma_2 ** 2 - mu_2 ** 2) / p_1 + sigma_2 ** 2 + mu_2 ** 2 - mu_1 ** 2)
        if sigma_1_squared < 0:
            # Validity check 3: check for upcoming complex numbers.
            break

        sigma_1 = sigma_1_squared ** 0.5

        # Adjust the guess for mu_2, Equation (27).
        if (1 - p_1) < 1e-4:
            # Validity check 5: break to prevent divide-by-zero.
            break

        a_1_squared = 6 * sigma_2 ** 4 + (m_4 - p_1 * (3 * sigma_1 ** 4 + 6 * sigma_1 ** 2 * mu_1 ** 2 + mu_1 ** 4)
                                          ) / (1 - p_1)
        if a_1_squared < 0:
            # Validity check 6: break to avoid taking the square root of negative number.
            break

        a_1 = a_1_squared ** 0.5
        mu_2_squared = a_1 - 3 * sigma_2 ** 2

        # Validity check 7: break to avoid complex numbers.
        # Todo: Avoid Numba object mode.
        # Numba does not support numpy.iscomplex. This creates an overhead.
        with objmode(mu_2_squared_is_complex="boolean"):
            mu_2_squared_is_complex = bool(np.iscomplex(mu_2_squared))
        if mu_2_squared_is_complex or mu_2_squared < 0:
            break

        mu_2 = mu_2_squared ** 0.5
        # Adjust guess for p_1, Equation (28, 29).
        a_2 = 15 * sigma_1 ** 4 * mu_1 + 10 * sigma_1 ** 2 * mu_1 ** 3 + mu_1 ** 5
        b_2 = 15 * sigma_2 ** 4 * mu_2 + 10 * sigma_2 ** 2 * mu_2 ** 3 + mu_2 ** 5
        if (a_2 - b_2) == 0:
            # Validity check 8: break to prevent divide-by-zero.
            break

        p_1 = (m_5 - b_2) / (a_2 - b_2)
        if (p_1 < 0) or (p_1 > 1):
            # Validity check 9: p_1 value must be between 0 and 1.
            break

        # Add all new parameter estimates to the return list if no break has occurred before now.
        param_list = np.array([mu_1, mu_2, sigma_1, sigma_2, p_1], dtype=float64)

        # We only want this to execute once at most, so call a final break if one hasn't been called yet.
        break

    return param_list



# pylint: disable=missing-module-docstring
import numpy as np
from famlafl.online_portfolio_selection.base import OLPS


class RMR(OLPS):
    """
    This class implements the Confidence Weighted Mean Reversion strategy. It is reproduced with
    modification from the following paper:
    `D. Huang, J. Zhou, B. Li, S. C. H. Hoi and S. Zhou, "Robust Median Reversion Strategy for
    Online Portfolio Selection," in IEEE Transactions on Knowledge and Data Engineering, vol. 28,
    no. 9, pp. 2480-2493, 1 Sept. 2016. <https://www.ijcai.org/Proceedings/13/Papers/296.pdf>`_

    Robust Median Reversion uses a L1-median of historical prices to predict the next time's
    price relative returns. The new weights is then regularized to retain previous portfolio
    information but also seeks to maximize returns from the predicted window.
    """

    def __init__(self, epsilon, n_iteration, window, tau=0.001):
        """
        Initializes Robust Median Reversion with the given epsilon, n_iteration, window, and tau values.

        :param epsilon: (float) Reversion threshold with range [1, inf). Values of [15, 25] had the
                                highest returns for the original dataset provided by the authors.
        :param n_iteration: (int) Maximum number of iterations with range [2, inf). Iteration of 200
                                  produced a adequate balance between performance and computational
                                  time for the strategy.
        :param window: (int) Size of window with range [2, inf). Typically window of 2, 7, or 21
                             produced the highest results. This parameter depends on the underlying
                             data's price movements and mean reversion tendencies.
        :param tau: (float) Toleration level with range [0, 1). It is suggested to keep the toleration
                            level at 0.001 for computational efficiency.
        """
        self.epsilon = epsilon
        self.n_iteration = n_iteration
        self.window = window
        self.tau = tau
        self.np_asset_prices = None  # (np.array) Asset prices converted to np.array.
        super().__init__()

    def _initialize(self, asset_prices, weights, resample_by):
        """
        Initializes the important variables for the object.

        :param asset_prices: (pd.DataFrame) Historical asset prices.
        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :param resample_by: (str) Specifies how to resample the prices. 'D' for Day, 'W' for Week,
                                 'M' for Month. The inputs are based on pandas' resample method.
        """
        super(RMR, self)._initialize(asset_prices, weights, resample_by)

        # Check that epsilon value is correct.
        if self.epsilon < 1:
            raise ValueError("Epsilon values must be greater than 1.")

        # Check that the n_iteration is an integer.
        if not isinstance(self.n_iteration, int):
            raise ValueError("Number of iterations must be an integer.")

        # Check that the number of iterations is greater or equal to 2.
        if self.n_iteration < 2:
            raise ValueError("Number of iterations must be greater or equal to 2.")

        # Check that the window value is an integer.
        if not isinstance(self.window, int):
            raise ValueError("Window must be an integer.")

        # Check that the window value is greater or equal to 2.
        if self.window < 2:
            raise ValueError("Window must be greater or equal to 2.")

        # Check that tau is within range of [0, 1).
        if self.tau < 0 or self.tau >= 1:
            raise ValueError("Tau must be greater or equal to 0 and less than 1.")

        self.np_asset_prices = np.array(self.asset_prices)

    def _update_weight(self, time):
        """
        Predicts the next time's portfolio weight.

        :param time: (int) Current time period.
        :return: (np.array) Predicted weights.
        """
        # Until the relative time window, return original weights.
        if time < self.window - 1:
            return self.weights

        # Set the current predicted relatives value.
        current_prediction = self._calculate_predicted_relatives(time)

        # Set the deviation from the mean of current prediction.
        predicted_deviation = current_prediction - np.ones(self.number_of_assets) * np.mean(
            current_prediction)

        # Calculate alpha, the lagrangian multiplier.
        norm2 = np.linalg.norm(predicted_deviation, ord=1) ** 2

        # If norm2 is zero, return previous weights.
        if norm2 == 0:
            return self.weights
        alpha = np.minimum(0, (current_prediction * self.weights - self.epsilon) / norm2)

        # Update new weights.
        new_weights = self.weights - alpha * predicted_deviation

        # Project to simplex domain.
        new_weights = self._simplex_projection(new_weights)

        return new_weights

    def _calculate_predicted_relatives(self, time):
        # pylint: disable=unsubscriptable-object
        """
        Calculates the predicted relatives using l1 median.

        :param time: (int) Current time.
        :return: (np.array) Predicted relatives using l1 median.
        """
        # Calculate the L1 median of the price window.
        price_window = self.np_asset_prices[time - self.window + 1:time + 1]
        curr_prediction = np.median(price_window, axis=0)

        # Iterate until the maximum iteration allowed.
        for _ in range(self.n_iteration - 1):
            prev_prediction = curr_prediction
            # Transform mu according the Modified Weiszfeld Algorithm
            curr_prediction = self._transform(curr_prediction, price_window)

            # If null value or condition is satisfied, break.
            if curr_prediction.size == 0 or np.linalg.norm(prev_prediction - curr_prediction, ord=1) \
                    <= self.tau * np.linalg.norm(curr_prediction, ord=1):
                curr_prediction = prev_prediction
                break

        # Divide by the current time's price.
        predicted_relatives = curr_prediction / price_window[-1]

        return predicted_relatives

    @staticmethod
    def _transform(old_mu, price_window):
        """
        Calculates L1 median approximation by using the Modified Weiszfeld Algorithm.

        :param old_mu: (np.array) Current value of the predicted median value.
        :param price_window: (np.array) A window of prices provided by the user.
        :return: (np.array) New updated l1 median approximation.
        """
        # Calculate the difference set.
        diff = price_window - old_mu

        # Remove rows with all zeros.
        non_mu = diff[~np.all(diff == 0, axis=1)]

        # Edge case for identical price windows.
        if non_mu.shape[0] == 0:
            return non_mu

        # Number of zeros.
        n_zero = diff.shape[0] - non_mu.shape[0]

        # Calculate eta.
        eta = 0 if n_zero == 0 else 1

        # Calculate l1 norm of non_mu.
        l1_norm = np.linalg.norm(non_mu, ord=1, axis=1)

        # Calculate tilde.
        tilde = 1 / np.sum(1 / l1_norm) * np.sum(np.divide(non_mu.T, l1_norm), axis=1)

        # Calculate gamma.
        gamma = np.linalg.norm(
            np.sum(np.apply_along_axis(lambda x: x / np.linalg.norm(x, ord=1), 1, non_mu), axis=0),
            ord=1)

        # Calculate next_mu value.
        with np.errstate(invalid='ignore'):
            next_mu = np.maximum(0, 1 - eta / gamma) * tilde + np.minimum(1, eta / gamma) * old_mu
        return tilde if eta == 0 else next_mu



"""
Inter-bar feature generator which uses trades data and bars index to calculate inter-bar features
"""

import pandas as pd
import numpy as np
from famlafl.microstructural_features.entropy import get_shannon_entropy, get_plug_in_entropy, get_lempel_ziv_entropy, \
    get_konto_entropy
from famlafl.microstructural_features.encoding import encode_array
from famlafl.microstructural_features.second_generation import get_trades_based_kyle_lambda, \
    get_trades_based_amihud_lambda, get_trades_based_hasbrouck_lambda
from famlafl.microstructural_features.misc import get_avg_tick_size, vwap
from famlafl.microstructural_features.encoding import encode_tick_rule_array
from famlafl.util.misc import crop_data_frame_in_batches


# pylint: disable=too-many-instance-attributes

class MicrostructuralFeaturesGenerator:
    """
    Class which is used to generate inter-bar features when bars are already compressed.

    :param trades_input: (str or pd.DataFrame) Path to the csv file or Pandas DataFrame containing raw tick data
                                               in the format[date_time, price, volume]
    :param tick_num_series: (pd.Series) Series of tick number where bar was formed.
    :param batch_size: (int) Number of rows to read in from the csv, per batch.
    :param volume_encoding: (dict) Dictionary of encoding scheme for trades size used to calculate entropy on encoded messages
    :param pct_encoding: (dict) Dictionary of encoding scheme for log returns used to calculate entropy on encoded messages

    """

    def __init__(self, trades_input: (str, pd.DataFrame), tick_num_series: pd.Series, batch_size: int = 2e7,
                 volume_encoding: dict = None, pct_encoding: dict = None):
        """
        Constructor

        :param trades_input: (str or pd.DataFrame) Path to the csv file or Pandas DataFrame containing raw tick data
                                                   in the format[date_time, price, volume]
        :param tick_num_series: (pd.Series) Series of tick number where bar was formed.
        :param batch_size: (int) Number of rows to read in from the csv, per batch.
        :param volume_encoding: (dict) Dictionary of encoding scheme for trades size used to calculate entropy on encoded messages
        :param pct_encoding: (dict) Dictionary of encoding scheme for log returns used to calculate entropy on encoded messages
        """

        if isinstance(trades_input, str):
            self.generator_object = pd.read_csv(trades_input, chunksize=batch_size, parse_dates=[0])
            # Read in the first row & assert format
            first_row = pd.read_csv(trades_input, nrows=1)
            self._assert_csv(first_row)
        elif isinstance(trades_input, pd.DataFrame):
            self.generator_object = crop_data_frame_in_batches(trades_input, batch_size)
        else:
            raise ValueError('trades_input is neither string(path to a csv file) nor pd.DataFrame')

        # Base properties
        self.tick_num_generator = iter(tick_num_series)
        self.current_bar_tick_num = self.tick_num_generator.__next__()

        # Cache properties
        self.price_diff = []
        self.trade_size = []
        self.tick_rule = []
        self.dollar_size = []
        self.log_ret = []

        # Entropy properties
        self.volume_encoding = volume_encoding
        self.pct_encoding = pct_encoding
        self.entropy_types = ['shannon', 'plug_in', 'lempel_ziv', 'konto']

        # Batch_run properties
        self.prev_price = None
        self.prev_tick_rule = 0
        self.tick_num = 0

    def get_features(self, verbose=True, to_csv=False, output_path=None):
        """
        Reads a csv file of ticks or pd.DataFrame in batches and then constructs corresponding microstructural intra-bar features:
        average tick size, tick rule sum, VWAP, Kyle lambda, Amihud lambda, Hasbrouck lambda, tick/volume/pct Shannon, Lempel-Ziv,
        Plug-in entropies if corresponding mapping dictionaries are provided (self.volume_encoding, self.pct_encoding).
        The csv file must have only 3 columns: date_time, price, & volume.

        :param verbose: (bool) Flag whether to print message on each processed batch or not
        :param to_csv: (bool) Flag for writing the results of bars generation to local csv file, or to in-memory DataFrame
        :param output_path: (bool) Path to results file, if to_csv = True
        :return: (DataFrame or None) Microstructural features for bar index
        """

        if to_csv is True:
            header = True  # if to_csv is True, header should be written on the first batch only
            open(output_path, 'w').close()  # Clean output csv file

        # Read csv in batches
        count = 0
        final_bars = []
        cols = ['date_time', 'avg_tick_size', 'tick_rule_sum', 'vwap', 'kyle_lambda', 'kyle_lambda_t_value',
                'amihud_lambda', 'amihud_lambda_t_value', 'hasbrouck_lambda', 'hasbrouck_lambda_t_value']

        # Entropy features columns
        for en_type in self.entropy_types:
            cols += ['tick_rule_entropy_' + en_type]

        if self.volume_encoding is not None:
            for en_type in self.entropy_types:
                cols += ['volume_entropy_' + en_type]

        if self.pct_encoding is not None:
            for en_type in self.entropy_types:
                cols += ['pct_entropy_' + en_type]

        for batch in self.generator_object:
            if verbose:  # pragma: no cover
                print('Batch number:', count)

            list_bars, stop_flag = self._extract_bars(data=batch)

            if to_csv is True:
                pd.DataFrame(list_bars, columns=cols).to_csv(output_path, header=header, index=False, mode='a')
                header = False
            else:
                # Append to bars list
                final_bars += list_bars
            count += 1

            # End of bar index, no need to calculate further
            if stop_flag is True:
                break

        # Return a DataFrame
        if final_bars:
            bars_df = pd.DataFrame(final_bars, columns=cols)
            return bars_df

        # Processed DataFrame is stored in .csv file, return None
        return None

    def _reset_cache(self):
        """
        Reset price_diff, trade_size, tick_rule, log_ret arrays to empty when bar is formed and features are
        calculated

        :return: None
        """
        self.price_diff = []
        self.trade_size = []
        self.tick_rule = []
        self.dollar_size = []
        self.log_ret = []

    def _extract_bars(self, data):
        """
        For loop which calculates features for formed bars using trades data

        :param data: (tuple) Contains 3 columns - date_time, price, and volume.
        """

        # Iterate over rows
        list_bars = []

        for row in data.values:
            # Set variables
            date_time = row[0]
            price = float(row[1])
            volume = row[2]
            dollar_value = price * volume
            signed_tick = self._apply_tick_rule(price)

            self.tick_num += 1

            # Derivative variables
            price_diff = self._get_price_diff(price)
            log_ret = self._get_log_ret(price)

            self.price_diff.append(price_diff)
            self.trade_size.append(volume)
            self.tick_rule.append(signed_tick)
            self.dollar_size.append(dollar_value)
            self.log_ret.append(log_ret)

            self.prev_price = price

            # If date_time reached bar index
            if self.tick_num >= self.current_bar_tick_num:
                self._get_bar_features(date_time, list_bars)

                # Take the next tick number
                try:
                    self.current_bar_tick_num = self.tick_num_generator.__next__()
                except StopIteration:
                    return list_bars, True  # Looped through all bar index
                # Reset cache
                self._reset_cache()
        return list_bars, False

    def _get_bar_features(self, date_time: pd.Timestamp, list_bars: list) -> list:
        """
        Calculate inter-bar features: lambdas, entropies, avg_tick_size, vwap

        :param date_time: (pd.Timestamp) When bar was formed
        :param list_bars: (list) Previously formed bars
        :return: (list) Inter-bar features
        """
        features = [date_time]

        # Tick rule sum, avg tick size, VWAP
        features.append(get_avg_tick_size(self.trade_size))
        features.append(sum(self.tick_rule))
        features.append(vwap(self.dollar_size, self.trade_size))

        # Lambdas
        features.extend(get_trades_based_kyle_lambda(self.price_diff, self.trade_size, self.tick_rule))  # Kyle lambda
        features.extend(get_trades_based_amihud_lambda(self.log_ret, self.dollar_size))  # Amihud lambda
        features.extend(
            get_trades_based_hasbrouck_lambda(self.log_ret, self.dollar_size, self.tick_rule))  # Hasbrouck lambda

        # Entropy features
        encoded_tick_rule_message = encode_tick_rule_array(self.tick_rule)
        features.append(get_shannon_entropy(encoded_tick_rule_message))
        features.append(get_plug_in_entropy(encoded_tick_rule_message))
        features.append(get_lempel_ziv_entropy(encoded_tick_rule_message))
        features.append(get_konto_entropy(encoded_tick_rule_message))

        if self.volume_encoding is not None:
            message = encode_array(self.trade_size, self.volume_encoding)
            features.append(get_shannon_entropy(message))
            features.append(get_plug_in_entropy(message))
            features.append(get_lempel_ziv_entropy(message))
            features.append(get_konto_entropy(message))

        if self.pct_encoding is not None:
            message = encode_array(self.log_ret, self.pct_encoding)
            features.append(get_shannon_entropy(message))
            features.append(get_plug_in_entropy(message))
            features.append(get_lempel_ziv_entropy(message))
            features.append(get_konto_entropy(message))

        list_bars.append(features)

    def _apply_tick_rule(self, price: float) -> int:
        """
        Advances in Financial Machine Learning, page 29.

        Applies the tick rule

        :param price: (float) Price at time t
        :return: (int) The signed tick
        """
        if self.prev_price is not None:
            tick_diff = price - self.prev_price
        else:
            tick_diff = 0

        if tick_diff != 0:
            signed_tick = np.sign(tick_diff)
            self.prev_tick_rule = signed_tick
        else:
            signed_tick = self.prev_tick_rule

        return signed_tick

    def _get_price_diff(self, price: float) -> float:
        """
        Get price difference between ticks

        :param price: (float) Price at time t
        :return: (float) Price difference
        """
        if self.prev_price is not None:
            price_diff = price - self.prev_price
        else:
            price_diff = 0  # First diff is assumed 0
        return price_diff

    def _get_log_ret(self, price: float) -> float:
        """
        Get log return between ticks

        :param price: (float) Price at time t
        :return: (float) Log return
        """
        if self.prev_price is not None:
            log_ret = np.log(price / self.prev_price)
        else:
            log_ret = 0  # First return is assumed 0
        return log_ret

    @staticmethod
    def _assert_csv(test_batch):
        """
        Tests that the csv file read has the format: date_time, price, and volume.
        If not then the user needs to create such a file. This format is in place to remove any unwanted overhead.

        :param test_batch: (pd.DataFrame) the first row of the dataset.
        :return: (None)
        """
        assert test_batch.shape[1] == 3, 'Must have only 3 columns in csv: date_time, price, & volume.'
        assert isinstance(test_batch.iloc[0, 1], float), 'price column in csv not float.'
        assert not isinstance(test_batch.iloc[0, 2], str), 'volume column in csv not int or float.'

        try:
            pd.to_datetime(test_batch.iloc[0, 0])
        except ValueError as e:
            raise ValueError(f'csv file, column 0, not a date time format: {test_batch.iloc[0, 0]}') from e



# pylint: disable=missing-module-docstring
import unittest
import os
import pandas as pd
from famlafl.portfolio_optimization.risk_metrics import RiskMetrics


class TestRiskMetrics(unittest.TestCase):
    """
    Tests different risk metrics calculation from the RiskMetrics class.
    """

    def setUp(self):
        """
        Set the file path for the tick data csv.
        """
        project_path = os.path.dirname(__file__)
        data_path = project_path + '/test_data/stock_prices.csv'
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date")

    def test_variance_calculation(self):
        """
        Test the calculation of variance.
        """

        weights = [1] * self.data.shape[1]
        variance = RiskMetrics().calculate_variance(self.data.cov(), weights)
        assert isinstance(variance, float)

    def test_value_at_risk_calculation(self):
        """
        Test the calculation of value at risk.
        """

        test_returns = self.data.iloc[:, 0].values
        value_at_risk = RiskMetrics().calculate_value_at_risk(test_returns)
        assert isinstance(value_at_risk, float)

    def test_expected_shortfall_calculation(self):
        """
        Test the calculation of expected shortfall.
        """

        test_returns = self.data.iloc[:, 0].values
        expected_shortfall = RiskMetrics().calculate_expected_shortfall(test_returns)
        assert isinstance(expected_shortfall, float)

    def test_conditional_drawdown_calculation(self):
        """
        Test the calculation of conditional drawdown at risk.
        """

        test_returns = self.data.iloc[:, 0].values
        conditional_drawdown = RiskMetrics().calculate_conditional_drawdown_risk(test_returns)
        assert isinstance(conditional_drawdown, float)

    def test_value_at_risk_for_dataframe(self):
        """
        Test the calculation of value at risk.
        """

        test_returns = pd.DataFrame(self.data.iloc[:, 0])
        value_at_risk = RiskMetrics().calculate_value_at_risk(test_returns)
        assert isinstance(value_at_risk, float)

    def test_expected_shortfall_for_dataframe(self):
        """
        Test the calculation of expected shortfall.
        """

        test_returns = pd.DataFrame(self.data.iloc[:, 0])
        expected_shortfall = RiskMetrics().calculate_expected_shortfall(test_returns)
        assert isinstance(expected_shortfall, float)

    def test_conditional_drawdown_for_dataframe(self):
        """
        Test the calculation of conditional drawdown at risk.
        """

        test_returns = pd.DataFrame(self.data.iloc[:, 0])
        conditional_drawdown = RiskMetrics().calculate_conditional_drawdown_risk(test_returns)
        assert isinstance(conditional_drawdown, float)



# pylint: disable=missing-module-docstring
# pylint: disable=protected-access

import unittest
import os
import numpy as np
import pandas as pd
from famlafl.labeling.matrix_flags import MatrixFlagLabels


class TestMatrixFlagLabels(unittest.TestCase):
    """
    Tests for the matrix flags labeling method.
    """

    def setUp(self):
        """
        Set the file path for data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/close_df.csv'
        self.close = pd.read_csv(self.path, index_col='date', parse_dates=True)

    def test_init(self):
        """
        Tests that exceptions are raised correctly during initialization of the class if inputs are wrong.
        """
        # Length of data less than 10.
        close = self.close['spx']
        with self.assertRaises(Exception):
            MatrixFlagLabels(close[0:7], window=30)
        # Window less than 10.
        with self.assertRaises(Exception):
            MatrixFlagLabels(close[0:30], window=5)
        # Window greater than len(data).
        with self.assertRaises(Exception):
            MatrixFlagLabels(close[0:30], window=50)
        # Data is a DataFrame, not Series.
        with self.assertRaises(Exception):
            MatrixFlagLabels(self.close, window=60)

    def test_set_template(self):
        """
        Tests for user setting a new template. Also verifies that exception is raised for invalid template formats.
        """
        new_template = pd.DataFrame(np.random.randint(-3, 3, size=(10, 10)))
        flags = MatrixFlagLabels(self.close['spx'], window=100)
        flags.set_template(new_template)  # Correct template

        bad_shape = pd.DataFrame(np.random.randint(-3, 3, size=(9, 10)))  # Not 10 by 10
        with self.assertRaises(Exception):
            flags.set_template(bad_shape)

        nan_template = new_template.copy()
        nan_template.iloc[2, 3] = np.nan  # Has NaN
        with self.assertRaises(Exception):
            flags.set_template(nan_template)

    def test_transform_data(self):
        """
        Tests that the transform_data method gives the correct 10 by 10 matrix.
        """
        data = self.close['spx'][3:103]
        flags = MatrixFlagLabels(data, window=30)
        test1 = flags._transform_data(row_num=32, window=30)
        test2 = flags._transform_data(row_num=100, window=90)

        test1_actual = pd.DataFrame([[0, 1. / 3, 0, 0, 1. / 3, 1. / 3, 0, 0, 0, 0],
                                     [0, 2. / 3, 0, 0, 1. / 3, 0, 0, 0, 0, 0],
                                     [0, 0, 1. / 3, 2. / 3, 0, 0, 0, 0, 0, 0],
                                     [1. / 3, 0, 0, 0, 0, 1. / 3, 0, 0, 0, 0],
                                     [1. / 3, 0, 1. / 3, 1. / 3, 1. / 3, 0, 0, 0, 0, 0],
                                     [1. / 3, 0, 1. / 3, 0, 0, 0, 0, 0, 1. / 3, 0],
                                     [0, 0, 0, 0, 0, 0, 0, 1. / 3, 1. / 3, 1. / 3],
                                     [0, 0, 0, 0, 0, 0, 0, 1. / 3, 1. / 3, 1. / 3],
                                     [0, 0, 0, 0, 0, 0, 2. / 3, 0, 0, 1. / 3],
                                     [0, 0, 0, 0, 0, 1. / 3, 1. / 3, 1. / 3, 0, 0]])
        test2_column_actual = pd.Series([0, 0, 0, 0, 0, 0, 0, 1. / 3, 1. / 9, 5. / 9])
        pd.testing.assert_frame_equal(test1, test1_actual)
        pd.testing.assert_series_equal(test2[1], test2_column_actual, check_names=False)  # Check 1 column

        # Assert that all columns sum to 1.
        self.assertTrue(test2.sum(axis=0).values.all() == np.array([1.] * 10).all())

    def test_apply_template(self):
        """
        Tests for the apply template to matrix. A matrix is used which satisfies the constraints of transform_data.
        Then, the template is changed, and the applied to the same matrix.
        """
        matrix = pd.DataFrame([[0, 1. / 3, 0, 0, 1. / 3, 1. / 3, 0, 0, 0, 0],
                               [0, 2. / 3, 0, 0, 1. / 3, 0, 0, 0, 0, 0],
                               [0, 0, 1. / 3, 2. / 3, 0, 0, 0, 0, 0, 0],
                               [1. / 3, 0, 0, 0, 0, 1. / 3, 0, 0, 0, 0],
                               [1. / 3, 0, 1. / 3, 1. / 3, 1. / 3, 0, 0, 0, 0, 0],
                               [1. / 3, 0, 1. / 3, 0, 0, 0, 0, 0, 1. / 3, 0],
                               [0, 0, 0, 0, 0, 0, 0, 1. / 3, 1. / 3, 1. / 3],
                               [0, 0, 0, 0, 0, 0, 0, 1. / 3, 1. / 3, 1. / 3],
                               [0, 0, 0, 0, 0, 0, 2. / 3, 0, 0, 1. / 3],
                               [0, 0, 0, 0, 0, 1. / 3, 1. / 3, 1. / 3, 0, 0]])

        flags = MatrixFlagLabels(self.close['spx'], window=100, template_name='leigh_bull')
        test3 = flags._apply_template_to_matrix(matrix, flags.template)
        self.assertAlmostEqual(test3, 2.6666666667)

        # Replace the template with another bull template proposed in Cervello-Royo's paper.
        new_template = pd.DataFrame([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                     [0, 0, 0, 0, -1, -1, -1, -1, -1, -1],
                                     [0, 0, 0, -1, -2, -2, -2, -2, -2, -2],
                                     [0, 0, -1, -3, -3, -3, -3, -3, -3, -3],
                                     [0, -1, -3, -5, -5, -5, -5, -5, -5, -5],
                                     [0, -1, -5, -5, -5, -5, -5, -5, -5, -5],
                                     [0, -1, -5, -5, -5, -5, -5, -5, -5, -5],
                                     [5, -1, -5, -5, -5, -5, -5, -5, -5, -5]])

        flags.set_template(new_template)
        test4 = flags._apply_template_to_matrix(matrix, flags.template)
        self.assertAlmostEqual(test4, -22.666666667)

    def test_apply_labeling(self):
        """
        Test for the function the users would actually use, for creating full labels from matrix.
        """
        data = self.close['spx'][0:100]
        flags = MatrixFlagLabels(prices=data, window=60, template_name='leigh_bull')
        test5 = flags.apply_labeling_matrix()

        # Verify that the output has 40 (100 - 60) rows.
        self.assertTrue(len(test5) == 40)

        # Check values.
        test5_subset = test5[10:20]
        test5_subset_actual = pd.Series([-0.91666667, -1.25, -1.75, -2.16666667, -2.66666667, -2.66666667, -2.75,
                                         -2.5, -2.58333333, -2.66666667], index=test5_subset.index)
        pd.testing.assert_series_equal(test5_subset, test5_subset_actual)

    def test_threshold(self):
        """
        Tests for when threshold is desired.
        """
        data = self.close['spx'][425:500]
        idx = data[60:].index
        flags = MatrixFlagLabels(prices=data, window=60, template_name='leigh_bull')
        test5 = flags.apply_labeling_matrix(threshold=4)
        test6 = flags.apply_labeling_matrix(threshold=-1.5)
        test5_actual = pd.Series([-1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], index=idx)
        test6_actual = pd.Series([1]*15, index=idx)
        pd.testing.assert_series_equal(test5, test5_actual)
        pd.testing.assert_series_equal(test6, test6_actual)

    def test_template_init(self):
        """
        Checks that other templates are given correctly.
        """
        close = self.close['spx']
        leigh_bull = pd.DataFrame([[.5, 0, -1, -1, -1, -1, -1, -1, -1, 0],
                                   [1, 0.5, 0, -0.5, -1, -1, -1, -1, -0.5, 0],
                                   [1, 1, 0.5, 0, -0.5, -0.5, -0.5, -0.5, 0, 0.5],
                                   [0.5, 1, 1, 0.5, 0, -0.5, -0.5, -0.5, 0, 1],
                                   [0, 0.5, 1, 1, 0.5, 0, 0, 0, 0.5, 1],
                                   [0, 0, 0.5, 1, 1, 0.5, 0, 0, 1, 1],
                                   [-0.5, 0, 0, 0.5, 1, 1, 0.5, 0.5, 1, 1],
                                   [-0.5, -1, 0, 0, 0.5, 1, 1, 1, 1, 0],
                                   [-1, -1, -1, -0.5, 0, 0.5, 1, 1, 0, -2],
                                   [-1, -1, -1, -1, -1, 0, 0.5, 0.5, -2, -2.5]])
        leigh_bear = pd.DataFrame(np.flip(np.array(leigh_bull), axis=0))
        cervelloroyo_bull = pd.DataFrame([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                          [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                          [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                          [0, 0, 0, 0, -1, -1, -1, -1, -1, -1],
                                          [0, 0, 0, -1, -2, -2, -2, -2, -2, -2],
                                          [0, 0, -1, -3, -3, -3, -3, -3, -3, -3],
                                          [0, -1, -3, -5, -5, -5, -5, -5, -5, -5],
                                          [0, -1, -5, -5, -5, -5, -5, -5, -5, -5],
                                          [0, -1, -5, -5, -5, -5, -5, -5, -5, -5],
                                          [5, -1, -5, -5, -5, -5, -5, -5, -5, -5]])
        cervelloroyo_bear = pd.DataFrame(np.flip(np.array(cervelloroyo_bull), axis=0))
        test7 = MatrixFlagLabels(close, 30, template_name='leigh_bear')
        test8 = MatrixFlagLabels(close, 30, template_name='cervelloroyo_bear')
        test9 = MatrixFlagLabels(close, 30, template_name='cervelloroyo_bull')
        pd.testing.assert_frame_equal(test7.template, leigh_bear)
        pd.testing.assert_frame_equal(test8.template, cervelloroyo_bear)
        pd.testing.assert_frame_equal(test9.template, cervelloroyo_bull)

        # Exception for invalid name.
        with self.assertRaises(Exception):
            MatrixFlagLabels(close[0:7], window=30, template_name='abcd')
        with self.assertRaises(Exception):
            test7._init_template('abcd')



# pylint: disable=missing-module-docstring

import unittest
import os
from datetime import datetime
import numpy as np
import pandas as pd


from famlafl.labeling.excess_over_median import excess_over_median


class TestLabelingOverMedian(unittest.TestCase):
    """
    Tests regarding labeling excess over median.
    """

    def setUp(self):
        """
        Set the file path for the sample data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/stock_prices.csv'
        self.data = pd.read_csv(self.path, index_col='Date')
        self.data.index = pd.to_datetime(self.data.index)

    def test_basic(self):
        """
        Test basic case for a small set with manually inputted results, with numerical and categorical outputs, with
        no resampling or forward looking labels.
        """
        cols = ['EEM', 'EWG', 'TIP']
        subset = self.data[cols].iloc[0:5]
        test1 = excess_over_median(subset, lag=False)
        test2 = excess_over_median(subset, binary=True, lag=False)
        test1_actual = pd.DataFrame([(np.nan, np.nan, np.nan), (0.0056216, -0.006201, 0), (-0.010485, 0, 0.019272),
                                     (0.006460, 0, -0.001054), (-0.000824, 0, 0.007678)],
                                    columns=cols, index=self.data[cols].iloc[0:5].index)
        test2_actual = test1_actual.apply(np.sign)

        pd.testing.assert_frame_equal(test1, test1_actual, rtol=0.001)
        pd.testing.assert_frame_equal(test2, test2_actual)

    def test_resample_period(self):
        """
        Test numerical and categorical with a resample period.
        """
        cols = ['EEM', 'EWG', 'TIP', 'EWJ']
        subset1 = self.data[cols].iloc[0:25]
        subset2 = self.data[cols].iloc[0:100]

        # Resample per business day. Should be same as no resampling at all (after removing Jan 21, since Python
        # considers that a business day even though it was MLK day during the year of the data, so no trading occurred).
        test3 = excess_over_median(subset1, resample_by='B', lag=False)
        test3.drop(datetime.strptime('2008-01-21', '%Y-%m-%d'), inplace=True)  # MLK day, 2008
        pd.testing.assert_frame_equal(test3, excess_over_median(subset1, lag=False))

        # Resample per week and month.
        test4 = excess_over_median(subset1, resample_by='W', lag=False)
        weekly_index = subset1.resample('W').last().index
        test5 = excess_over_median(subset2, resample_by='M', lag=False)
        monthly_index = subset2.resample('M').last().index
        test6 = excess_over_median(subset2, binary=True, resample_by='M', lag=False)  # Binary by month

        test4_actual = pd.DataFrame([(np.nan, np.nan, np.nan, np.nan),
                                     (0.019896, -0.009335, 0.007538, -0.007538),
                                     (-0.039458, -0.016603, 0.050073, 0.016603),
                                     (-0.006333, -0.060147, 0.015185, 0.006333),
                                     (0.009036, 0.010079, -0.039004, -0.009036),
                                     (-0.019975, 0.002612, 0.044291, -0.002612)],
                                    columns=cols, index=weekly_index)
        test5_actual = pd.DataFrame([(np.nan, np.nan, np.nan, np.nan),
                                     (0.018951, -0.008043, 0.008043, -0.015929),
                                     (-0.028116, 0.026522, 0.003355, -0.003355),
                                     (0.036035, -0.018217, -0.080735, 0.018217),
                                     (0.000479, 0.006833, -0.000479, -0.010351)],
                                    columns=cols, index=monthly_index)

        pd.testing.assert_frame_equal(test4, test4_actual, rtol=0.001)
        pd.testing.assert_frame_equal(test5, test5_actual, rtol=0.001)
        pd.testing.assert_frame_equal(test6, test5_actual.apply(np.sign))

    def test_forward(self):
        """
        Tests with lagged returns.
        """
        cols = ['EEM', 'EWG', 'TIP']
        subset = self.data[cols].iloc[0:5]
        subset2 = self.data[cols].iloc[0:100]
        monthly_index = subset2.resample('M').last().index

        test7 = excess_over_median(subset, lag=True)
        test8 = excess_over_median(subset, binary=True, lag=True)
        test9 = excess_over_median(subset2, resample_by='M', lag=True)
        test10 = excess_over_median(subset2, resample_by='M', binary=True, lag=True)

        test7_actual = pd.DataFrame([(0.0056216, -0.006201, 0), (-0.010485, 0, 0.019272), (0.006460, 0, -0.001054),
                                     (-0.000824, 0, 0.007678), (np.nan, np.nan, np.nan)],
                                    columns=cols, index=self.data[cols].iloc[0:5].index)
        test9_actual = pd.DataFrame([(0.010909, -0.016086, 0), (-0.031471, 0.023167, 0), (0.054252, 0, -0.062518),
                                     (0, 0.006354, -0.000958), (np.nan, np.nan, np.nan)],
                                    columns=cols, index=monthly_index)

        pd.testing.assert_frame_equal(test7, test7_actual, rtol=0.001)
        pd.testing.assert_frame_equal(test8, test7_actual.apply(np.sign))
        pd.testing.assert_frame_equal(test9, test9_actual, rtol=0.001)
        pd.testing.assert_frame_equal(test10, test9_actual.apply(np.sign))

    def test_nan(self):
        """
        Tests to check that NaN values in prices get ignored.
        """
        cols = ['EEM', 'EWG', 'TIP']
        subset = self.data[cols].iloc[0:5]
        with_nan = pd.concat([subset, pd.Series([np.nan]*5, name='nan', index=subset.index)], axis=1)
        test11 = excess_over_median(with_nan)
        test11.drop('nan', axis=1, inplace=True)
        pd.testing.assert_frame_equal(test11, excess_over_median(subset), rtol=0.001)



"""
Various functions for message encoding (quantile)
"""
import numpy as np


def encode_tick_rule_array(tick_rule_array: list) -> str:
    """
    Encode array of tick signs (-1, 1, 0)

    :param tick_rule_array: (list) Tick rules
    :return: (str) Encoded message
    """
    message = ''
    for element in tick_rule_array:
        if element == 1:
            message += 'a'
        elif element == -1:
            message += 'b'
        elif element == 0:
            message += 'c'
        else:
            raise ValueError('Unknown value for tick rule: {}'.format(element))
    return message


def _get_ascii_table() -> list:
    """
    Get all ASCII symbols

    :return: (list) ASCII symbols
    """
    # ASCII table consists of 256 characters
    table = []
    for i in range(256):
        table.append(chr(i))
    return table


def quantile_mapping(array: list, num_letters: int = 26) -> dict:
    """
    Generate dictionary of quantile-letters based on values from array and dictionary length (num_letters).

    :param array: (list) Values to split on quantiles
    :param num_letters: (int) Number of letters(quantiles) to encode
    :return: (dict) Dict of quantile-symbol
    """
    encoding_dict = {}
    ascii_table = _get_ascii_table()
    alphabet = ascii_table[:num_letters]
    for quant, letter in zip(np.linspace(0.01, 1, len(alphabet)), alphabet):
        encoding_dict[np.quantile(array, quant)] = letter
    return encoding_dict


def sigma_mapping(array: list, step: float = 0.01) -> dict:
    """
    Generate dictionary of sigma encoded letters based on values from array and discretization step.

    :param array: (list) Values to split on quantiles
    :param step: (float) Discretization step (sigma)
    :return: (dict) Dict of value-symbol
    """
    i = 0
    ascii_table = _get_ascii_table()
    encoding_dict = {}
    encoding_steps = np.arange(min(array), max(array), step)
    for element in encoding_steps:
        try:
            encoding_dict[element] = ascii_table[i]
        except IndexError:
            raise ValueError(
                'Length of dictionary ceil((max(arr) - min(arr)) / step = {} is more than ASCII table lenght)'.format(
                    len(encoding_steps)))
        i += 1
    return encoding_dict


def _find_nearest(array: list, value: float) -> float:
    """
    Find the nearest element from array to value.

    :param array: (list) Values
    :param value: (float) Value for which the nearest element needs to be found
    :return: (float) The nearest to the value element in array
    """
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]


def _get_letter_from_encoding(value: float, encoding_dict: dict) -> str:
    """
    Get letter for float/int value from encoding dict.

    :param value: (float/int) Value to use
    :param encoding_dict: (dict) Used dictionary
    :return: (str) Letter from encoding dict
    """
    return encoding_dict[_find_nearest(list(encoding_dict.keys()), value)]


def encode_array(array: list, encoding_dict: dict) -> str:
    """
    Encode array with strings using encoding dict, in case of multiple occurrences of the minimum values,
    the indices corresponding to the first occurrence are returned

    :param array: (list) Values to encode
    :param encoding_dict: (dict) Dict of quantile-symbol
    :return: (str) Encoded message
    """
    message = ''
    for element in array:
        message += _get_letter_from_encoding(element, encoding_dict)
    return message
